from emiproc.tests_utils.test_grids import regular_grid
from emiproc.tests_utils.test_inventories import inv_with_pnt_sources 
from emiproc.tests_utils import TEST_OUTPUTS_DIR

from emiproc.exports.rasters import export_raster_netcdf
from emiproc.utilities import Units

# Convert to the correct CRS
raster_inv = inv_with_pnt_sources.copy()
raster_inv.set_crs(regular_grid.crs)


def test_base_function():
    """Simply test that the function works with defaults"""

    export_raster_netcdf(
        raster_inv,
        TEST_OUTPUTS_DIR / "test_raster.nc",
        regular_grid,
        netcdf_attributes={},
    )

def test_group_categories():
    """Simply test that the function works with defaults"""

    export_raster_netcdf(
        raster_inv,
        TEST_OUTPUTS_DIR / "test_raster.nc",
        regular_grid,
        netcdf_attributes={},
        group_categories=True,
    )
        
def test_unit():

    export_raster_netcdf(
        raster_inv,
        TEST_OUTPUTS_DIR / "test_raster_kg_per_m2_per_s.nc",
        regular_grid,
        netcdf_attributes={},
        unit=Units.KG_PER_M2_PER_S,
    )


from __future__ import annotations
from enum import Enum, auto
from os import PathLike
from pathlib import Path
from emiproc.grids import LV95, SwissGrid
from emiproc.inventories import Category, Inventory, Substance
import pandas as pd
import geopandas as gpd
from shapely.geometry import Polygon, Point
from shapely.creation import polygons
import numpy as np
import rasterio


class PointSourceCorrection(Enum):
    """Enum for the point source correction method."""

    KEEP_RASTER_ONLY = auto()
    KEEP_POINT_SOURCE_ONLY_SCALED_TO_RASTER_TOTAL = auto()
    REMOVE_POINT_SOURCE_FROM_RASTER_TOTAL = auto()
    IS_ONLY_POINT_SOURCE = auto()


default_point_source_correction = {
    "eipro": PointSourceCorrection.REMOVE_POINT_SOURCE_FROM_RASTER_TOTAL,
    "eipzm": PointSourceCorrection.KEEP_POINT_SOURCE_ONLY_SCALED_TO_RASTER_TOTAL,
    "eipkv": PointSourceCorrection.KEEP_POINT_SOURCE_ONLY_SCALED_TO_RASTER_TOTAL,
    "eikla": PointSourceCorrection.KEEP_RASTER_ONLY,
    "eidep": PointSourceCorrection.KEEP_RASTER_ONLY,
    "eiprd": PointSourceCorrection.REMOVE_POINT_SOURCE_FROM_RASTER_TOTAL,
}


class SwissRasters(Inventory):
    """An inventory of Switzerland based on swiss rasters."""

    edge_size: int = 100
    grid: SwissGrid
    df_eipwp: gpd.GeoDataFrame
    df_emissions: pd.DataFrame
    emission: pd.Series

    def __init__(
        self,
        filepath_csv_totals: Path,
        filepath_point_sources: Path,
        rasters_dir: PathLike,
        rasters_str_dir: PathLike,
        requires_grid: bool = True,
        year: int = 2015,
        point_source_correction: dict[
            Category, PointSourceCorrection
        ] = default_point_source_correction,
    ) -> None:
        """Create a swiss raster inventory.

        :arg filepath_csv_totals: Csv file containing total emissions
            for each category/substance for different years.
        :arg filepath_point_sources: Excel file containing point sources.
            See in :py:func:`emiproc.inventories.swiss.read_prtr` for more details.
        :arg rasters_dir: The folder where the rasters are found.
        :arg rasters_str_dir: The folder where the rasters pro substance are found.
        :arg requires_grid: Whether the grid should be created as well.
            Creating the shapes for the swiss grid is quite expensive process.
            Most of the weights for remapping can be cached so if you
            have them generated already, set that to false.
        :arg year: The year of the inventory that should be used.
            This should be present in the `Emissions_CH.xlsx` file.
            The raster files are the same for all years. Only the scaling
            of the full raster pro substance changes.
        """
        super().__init__()

        self.year = year

        filepath_csv_totals = Path(filepath_csv_totals)

        # Emission data file
        if filepath_csv_totals.is_file():
            total_emission_file = filepath_csv_totals
        else:
            raise FileNotFoundError(
                f"Data path {filepath_csv_totals} is not an existing file or a folder."
            )

        # Load excel sheet with the total emissions (excluding point sources)
        df_emissions = pd.read_csv(total_emission_file, comment="#")

        # Add indexing column consisting of both grid and species' name
        df_emissions["cat_sub"] = (
            df_emissions["category"] + "_" + df_emissions["substance"]
        )
        df_emissions = df_emissions.set_index("cat_sub")
        self.df_emissions = df_emissions

        year_str = str(year)

        # Check if selected year is in dataset
        if year_str not in df_emissions.columns:
            raise ValueError(
                f"Selected {year=} not in dataset with columns={df_emissions.columns}."
            )

        emissions = df_emissions[year_str].copy()

        # List with substance
        substances = df_emissions["substance"].unique()

        # ---------------------------------------------------------------------
        # -- Emissions from point sources
        # ---------------------------------------------------------------------

        # Load data
        gdfs = read_prtr(filepath_point_sources, year, substances=substances)

        # Remove the total values in the rasters
        for cat, gdf in gdfs.items():

            if "CO2" in gdf.columns:
                # We need to split CO2 in CO2 and CO2_biog
                catsub_co2 = cat + "_CO2"
                catsub_biog = cat + "_CO2_biog"
                assert (
                    "CO2_biog" not in gdf.columns
                ), "Unexpected CO2_biog in the point sources"

                if (
                    point_source_correction[cat]
                    == PointSourceCorrection.IS_ONLY_POINT_SOURCE
                ):
                    biog_fracton = 0.0
                else:
                    # Get the biogenic fraction in the total emissions
                    # and apply it to the pointsources
                    biog_fracton = emissions.loc[catsub_biog] / (
                        emissions.loc[catsub_co2] + emissions.loc[catsub_biog]
                    )
                # Split the CO2 emissions in two
                base_col = gdf["CO2"].copy()
                gdf["CO2"] = base_col * (1.0 - biog_fracton)
                gdf["CO2_biog"] = base_col * biog_fracton

            totals = gdf.drop(columns=["geometry"]).sum(axis="rows")
            for sub in totals.index:
                catsub = cat + "_" + sub

                if cat not in point_source_correction:
                    raise ValueError(
                        f"Category {cat} with point source emissions not in point_source_correction dictionary."
                    )

                correction = point_source_correction[cat]

                if correction == PointSourceCorrection.KEEP_RASTER_ONLY:
                    # Remove the emissions of point sources
                    gdf[sub] = 0.0
                elif correction == PointSourceCorrection.IS_ONLY_POINT_SOURCE:
                    # Make sure the raster is empty
                    if catsub in emissions and emissions.loc[catsub] != 0:
                        raise ValueError(
                            f"Raster {catsub} is not empty for {correction}."
                        )
                    emissions.loc[catsub] = 0.0
                elif (
                    correction
                    == PointSourceCorrection.KEEP_POINT_SOURCE_ONLY_SCALED_TO_RASTER_TOTAL
                ):
                    # Scale the point sources to match
                    total = emissions.loc[catsub]
                    gdf[sub] = gdf[sub] / totals[sub] * total
                    # Remove the emissions of the raster
                    emissions.loc[catsub] = 0.0
                elif (
                    correction
                    == PointSourceCorrection.REMOVE_POINT_SOURCE_FROM_RASTER_TOTAL
                ):
                    if emissions.loc[catsub] < totals[sub]:
                        self.logger.warning(
                            f"Total emissions for {cat=} and {sub=} are negative."
                            f" point sources: {totals[sub]}, inventory total: {emissions.loc[catsub]}"
                            " Only the point sources will be used."
                        )
                        # If we have more emissions in the point sources than in the inventory
                        # We remove the rasters emissions
                        emissions.loc[catsub] = 0.0
                    else:
                        emissions.loc[catsub] -= totals[sub]
                else:
                    raise ValueError(f"Unknown correction {correction}")

        # ---------------------------------------------------------------------
        # -- Grids
        # ---------------------------------------------------------------------

        rasters_dir = Path(rasters_dir)

        # Grids that depend on substance (road transport)
        rasters_str_dir = Path(rasters_str_dir)
        str_rasters = [
            r
            for r in rasters_str_dir.glob("*.asc")
            # Don't include the tunnel specific grids as they are already included in the grids for road transport
            if "_tun" not in r.stem
        ]

        # Grids that do not depend on substance
        normal_rasters = [r for r in rasters_dir.glob("*.asc")]

        self.all_raster_files = normal_rasters + str_rasters

        self.raster_categories = [r.stem for r in normal_rasters] + [
            r.stem for r in str_rasters
        ]

        # List with Raster categories for which we have emissions
        raster_sub = emissions.index.tolist()
        rasters_w_emis = []

        evstr_subname_to_subname = {}
        for t in raster_sub:
            split = t.split("_")
            assert len(split) > 1
            cat = split[0]
            sub = "_".join(split[1:])
            if "evstr" in cat:
                # Grid for non-methane VOCs is named "evstr_nmvoc"
                subname = sub.lower()
                if subname == "voc":
                    subname = "nmvoc"
                rasters_w_emis.append(cat + "_" + subname)
                evstr_subname_to_subname[subname] = sub
            elif (
                cat in point_source_correction
                and point_source_correction[cat]
                == PointSourceCorrection.IS_ONLY_POINT_SOURCE
            ):
                # If the category is only point source, we don't need the raster
                pass
            else:
                rasters_w_emis.append(cat)
        # Remove duplicates
        rasters_w_emis = [*set(rasters_w_emis)]

        # Compare Raster categories of input emission file with Raster categories of grids
        # Raise error if the two don't agree
        if not sorted(self.raster_categories) == sorted(rasters_w_emis):
            missing_raster_files = [
                r for r in rasters_w_emis if r not in self.raster_categories
            ]
            missing_emissions_values = [
                r for r in self.raster_categories if r not in rasters_w_emis
            ]
            raise ValueError(
                "Raster categories of emission file don't match:"
                f"\nMissing raster files: {missing_raster_files}"
                f"\nMissing emissions values: {missing_emissions_values}"
            )

        # ---------------------------------------------------------------------
        # -- Emissions without point sources
        # ---------------------------------------------------------------------

        self.emission = emissions
        self.requires_grid = requires_grid

        # Grid on which the inventory is created
        self.grid = SwissGrid(
            "ch_emissions",
            nx=3600,
            ny=2400,
            xmin=2480000,
            ymin=1060000,
            dx=self.edge_size,
            dy=self.edge_size,
            crs=LV95,
        )

        xs = np.arange(
            self.grid.xmin,
            self.grid.xmin + self.grid.nx * self.grid.dx,
            step=self.grid.dx,
        )
        ys = np.arange(
            self.grid.ymin,
            self.grid.ymin + self.grid.ny * self.grid.dy,
            step=self.grid.dy,
        )

        mapping = {}

        # Loading Raster categories and assigning respective emissions
        for raster_file, category in zip(self.all_raster_files, self.raster_categories):
            _raster_array = self.load_raster(raster_file).reshape(-1)
            if "_" in category:
                split = category.split("_")
                cat = split[0]
                sub = "_".join(split[1:])
                sub_name = evstr_subname_to_subname[sub]
                idx = cat + "_" + sub_name
                total_emissions = emissions.loc[idx]
                # Normalize the array to ensure the factor will be the sum
                # Note: this is to ensure consistency if the data provider
                # change the df_emission values in the future but not the rasters
                _normalized_raster_array = _raster_array / _raster_array.sum()
                mapping[(cat, sub_name)] = _normalized_raster_array * total_emissions
            else:
                for sub in substances:
                    idx = category + "_" + sub
                    total_emissions = emissions.loc[idx]
                    if total_emissions > 0:
                        mapping[(category, sub)] = _raster_array * total_emissions

        if self.requires_grid:
            x_coords, y_coords = np.meshgrid(xs, ys[::-1])
            # Reshape to 1D
            x_coords = x_coords.flatten()
            y_coords = y_coords.flatten()
            dx = self.grid.dx
            dy = self.grid.dy
            coords = np.array(
                [
                    [x, y]
                    for x, y in zip(
                        [x_coords, x_coords, x_coords + dx, x_coords + dx],
                        [y_coords, y_coords + dy, y_coords + dy, y_coords],
                    )
                ]
            )
            coords = np.rollaxis(coords, -1, 0)
        self.gdf = gpd.GeoDataFrame(
            mapping,
            crs=LV95,
            # This vector is same as raster data reshaped using reshape(-1)
            geometry=(
                polygons(coords)
                if self.requires_grid
                else np.full(self.grid.nx * self.grid.ny, np.nan)
            ),
        )

        # Add point sources
        self.gdfs = gdfs

    def load_raster(self, raster_file: Path) -> np.ndarray:
        # Load and save as npy fast reading format
        if raster_file.with_suffix(".npy").exists():
            inventory_field = np.load(raster_file.with_suffix(".npy"))
        else:
            self.logger.info(f"Parsing {raster_file}")
            src = rasterio.open(raster_file).read(1)
            np.save(raster_file.with_suffix(".npy"), src)
            inventory_field = src
        return inventory_field


polluant_matching = {
    #'PCDD +PCDF (Dioxine + Furane) (als Teq)',
    #'Gesamtphosphor',
    #'Ethylenoxid',
    #'Gesamtstickstoff ',
    "Schwefeloxide (SOx/SO2)": "SO2",
    #'Halone',
    #'Arsen und Verbindungen (als As)',
    #'Dichlormethan (DCM)',
    #'Nonylphenolethoxylate (NP/NPEs) und verwandte Stoffe',
    "flüchtige organische Verbindungen ohne Methan (NMVOC)": "VOC",
    "Kohlenmonoxid (CO)": "CO",
    "Stickstoffoxide (NOx/NO2)": "NOx",
    #'Xylole (als BTEX) (a)',
    #'gesamter organischer Kohlenstoff (TOC) (als Gesamt-C oder CSB/3)',
    #'Toluol (als BTEX) (a)',
    "Kohlendioxid (CO2)": "CO2",
    #'Benzol (als BTEX) (a)',
    "Fluoride (als Gesamt-F)": "F-Gases",
    "Ammoniak (NH3)": "NH3",
    #'Chloride (als Gesamt-Cl)',
    #'Quecksilber und Verbindungen (als Hg)', 'Cyanide (als Gesamt-CN)',
    #'Chrom und Verbindungen (als Cr)',
    # 'Kupfer und Verbindungen (als Cu)',
    # 'Nickel und Verbindungen (als Ni)',
    # 'Blei und Verbindungen (als Blei)',
    # 'Cadmium und Verbindungen (als Cd)',
    # 'Zink und Verbindungen (als Zn)',
    "Feinstaub (PM10)": "PM10",
    # 'Fluor und anorganische Verbindungen (als HF)',
    # 'Chlor und anorganische Verbindungen (als HCl)',
    "Methan (CH4)": "CH4",
    # 'Halogenierte organische Verbindungen (als AOX)',
    # 'Tributylzinn und Verbindungen', 'Ethylbenzol (als BTEX) (a)',
    # 'Phenole (als Gesamt-C)',
    "Distickstoffoxid (N2O)": "N2O",
    # 'Trichlorethen',
    "Schwefelhexafluorid (SF6)": "SF6",
    # 'teilfluorierte Kohlenwasserstoffe (HFKW)',
    # 'Polychlorierte Biphenyle (PCB)', 'Diuron',
    # 'Tetrachlorethen (PER)', 'Naphthalin',
    # 'Polyzyklische aromatische Kohlenwasserstoffe (PAK) (b)',
    # '1,2-Dichlorethan (EDC)', 'Cyanwasserstoff (HCN) '
}


activities_to_categories = {
    # 1 - Energiesektor
    "1.a": "eipro",  # Mineralöl- und Gasraffinerien
    "1.b": "eipro",  # Vergasungs- und Verflüssigungsanlagen
    "1.c": "eipro",  # Wärmekraftwerke und andere Verbrennungsanlagen
    # 2 - Herstellung und Verarbeitung von Metallen
    "2.b": "eipro",
    "2.c.1": "eipro",
    "2.c.2": "eipro",
    "2.e.1": "eipro",
    "2.e.2": "eipro",
    "2.f": "eipro",
    # 3 - Mineralverarbeitende Industrie zement: eipzm=point source zement
    "3.c.1": "eipzm",  # c - Anlagen zur Herstellung von 1 - Zementklinkern in Drehrohröfen
    # 3 - Mineralverarbeitende Industrie others
    "3.e": "eipro",  # Glas
    "3.f": "eipro",  # Schmelzen mineralischer Stoffe
    "3.g": "eipro",  #  Herstellung von keramischen Erzeugnissen
    # 4 - Chemische Industrie
    "4.a.1": "eipro",
    "4.a.10": "eipro",
    "4.a.11": "eipro",
    "4.a.2": "eipro",
    "4.a.5": "eipro",
    "4.a.8": "eipro",
    "4.b.5": "eipro",
    "4.d": "eipro",
    "4.e": "eipro",
    "4.f": "eipro",
    # 5 - Abfall- und Abwasserbewirtschaftung
    # Punktquellen KVA (Kehrichtverbrennungsanlagen ) == Waste incinerators
    "5.a": "eipkv",  #  Anlagen zur Verbrennung, Pyrolyse, Verwertung, chemischen Behandlung, oder Deponierung von Sonderabfällen
    "5.b": "eipkv",  # Kehrichtverbrennungsanlagen für Siedlungsmüll
    "5.d": "eidep",  # Deponien
    "5.f": "eikla",  # Kommunale Abwasserbehandlungsanlagen
    "5.g": "eikla",  # Eigenständig betriebene Industrie¬abwasserbehandlungsanlagen
    # 6 - Be- und Verarbeitung von Papier und Holz
    "6.b": "eipro",
    # 8 - Tierische und pflanzliche Produkte aus dem Lebensmittel- und Getränkesektor
    "8.b.2": "eipro",
    "8.c": "eipro",
    # 9 - Sonstige Tätigkeiten
    "9.c": "eipro",
    "9.d": "eipro",
}


def read_prtr(
    prtr_file: PathLike, year: int, substances: list[Substance] | None = None
) -> dict[Category, gpd.GeoDataFrame]:
    """Read the PRTR file and return the gdfs.

    If you want to change the substances or the categories, you can do it
    by changing the dictionaries in the `emiproc.inventories.swiss` module.

    :arg prtr_file: Path to the PRTR file.
        This file must be downloaded from the Swiss PRTR website
        https://www.bafu.admin.ch/bafu/de/home/themen/chemikalien/zustand/schadstoffregister-swissprtr.html
        then go in the 'Datenpublikation' section and download the Excel file
    :arg year: The year of the data to use.
    :arg substances: List of substances to use.
        If None, all substances will be used.

    :returns: A dictionary with the categories as keys and the GeoDataFrames as values.
        Can be used to create the Inventory object.
    """

    df_prtr = pd.read_excel(prtr_file, skiprows=[0, 1, 3])

    substance_matching = {
        key: value
        for key, value in polluant_matching.items()
        if substances is None or value in substances
    }
    # Check that all substances are in the dictionary
    if substances is not None:
        for sub in substances:
            if sub not in substance_matching.values() and sub not in [
                # Compounds which are not in the point sources data
                "CO2_biog",
                "PM25",
            ]:
                raise ValueError(
                    f"Unkown substance `{sub}` not in the "
                    "`emiproc.inventories.swiss.polluant_matching` dictionary."
                )

    columns_of_interest = [
        "Year",
        "North coordinate (CH1903+)",
        "East coordinate (CH1903+)",
        "Facility",
        "Value",
        "Unit",
        "Pollutant_name",
        "Installation_main activity",
    ]

    mask_point_source = df_prtr["Source type"] == "Punktquelle"
    if year not in df_prtr["Year"].unique():
        raise ValueError(f"Year {year} not in the data.")
    mask_year = df_prtr["Year"] == year
    mask_has_value = df_prtr["Value"].notnull()
    mask_known_pollutants = df_prtr["Pollutant_name"].isin(substance_matching.keys())
    df_cleaned = df_prtr.loc[
        mask_point_source & mask_year & mask_has_value & mask_known_pollutants,
        columns_of_interest,
    ]

    # Correct the units to have kg year
    factors = {
        "t/a": 1e3,
        "kg/a": 1.0,
    }
    mask_corrected = pd.Series(0, index=df_cleaned.index, dtype=bool)
    for unit, factor in factors.items():
        mask = df_cleaned["Unit"] == unit
        mask_corrected = mask_corrected | mask
        df_cleaned.loc[mask, "Value"] *= factor
    if not mask_corrected.all():
        raise ValueError(
            f"Units not corrected for {df_cleaned.loc[~mask_corrected, 'Unit'].unique()}."
            " Fix the `emiproc.inventories.swiss.polluant_matching` dictionary."
        )
    # Drop the unit column
    df_cleaned = df_cleaned.drop(columns="Unit")
    # Rename the substances
    df_cleaned["Substance"] = df_cleaned["Pollutant_name"].replace(substance_matching)
    # Check the categories
    mask_matching = df_cleaned["Installation_main activity"].isin(
        activities_to_categories.keys()
    )
    if not mask_matching.all():
        raise ValueError(
            f"Missing categories for {df_cleaned.loc[~mask_matching, 'Installation_main activity'].unique()}"
            " Fix the `emiproc.inventories.swiss.activities_to_categories` dictionary."
        )
    df_cleaned["Category"] = df_cleaned["Installation_main activity"].replace(
        activities_to_categories
    )
    df_cleaned["x"] = df_cleaned["East coordinate (CH1903+)"]
    df_cleaned["y"] = df_cleaned["North coordinate (CH1903+)"]

    cols_for_emiproc = ["Category", "Substance", "Value", "x", "y"]
    df_emiproc = df_cleaned[cols_for_emiproc].copy()
    df_emiproc

    gdfs = {}
    for category in df_emiproc["Category"].unique():
        cat_df = df_emiproc.loc[df_emiproc["Category"] == category].copy()
        for substances in cat_df["Substance"].unique():
            mask_substance = cat_df["Substance"] == substances
            df_substance = cat_df.loc[mask_substance].copy()
            cat_df[substances] = df_substance["Value"]
        df_groupped = (
            cat_df.drop(columns=["Value", "Substance", "Category"])
            .groupby(["x", "y"])
            .sum()
            .reset_index()
        )
        gdfs[category] = gpd.GeoDataFrame(
            df_groupped.drop(columns=["x", "y"]),
            geometry=gpd.points_from_xy(df_groupped["x"], df_groupped["y"]),
            crs="LV95",
        )

    return gdfs



from pathlib import Path

import pytest

import emiproc

# get the path of the emiproc module
emiproc_path = Path(*emiproc.__path__)
test_utils_path = emiproc_path / "tests_utils"

# List all the python files in the tests_utils directory
modules = [f.stem for f in test_utils_path.glob("*.py")]


@pytest.mark.parametrize(
    "module",
    [(m) for m in modules],
    ids=modules,
)
def test_import_test_module(module):
    # Call import on the module
    exec(f"import emiproc.tests_utils.{module}")



import geopandas as gpd
from shapely.geometry import Point, Polygon

from emiproc.inventories import Inventory
from emiproc.inventories.utils import crop_with_shape
from emiproc.regrid import geoserie_intersection
from emiproc.tests_utils import WEIGHTS_DIR
from emiproc.tests_utils.test_inventories import inv_only_one_gdfs

serie = gpd.GeoSeries(
    [
        Polygon(((0, 0), (0, 1), (1, 1), (1, 0))),
        Polygon(((0, 1), (0, 2), (1, 2), (1, 1))),
        Polygon(((1, 0), (1, 1), (2, 1), (2, 0))),
        Polygon(((1, 1), (1, 2), (2, 2), (2, 1))),
        Polygon(((2, 1), (2, 2), (3, 2), (3, 1))),
    ]
)
triangle = Polygon(((0.5, 0.5), (1.5, 0.5), (1.5, 1.5)))

# Check the intersection
cropped, weights = geoserie_intersection(
    serie, triangle, keep_outside=True, drop_unused=False
)


inv = Inventory.from_gdf(
    gpd.GeoDataFrame(
        {
            ("adf", "CH4"): [i + 3 for i in range(len(serie))],
            ("adf", "CO2"): [i for i in range(len(serie))],
            ("liku", "CO2"): [i for i in range(len(serie))],
            ("test", "NH3"): [i + 1 for i in range(len(serie))],
        },
        geometry=serie,
    )
)

inv.gdf
inv_with_pnt_sources = inv.copy()
inv_with_pnt_sources.gdfs["blek"] = gpd.GeoDataFrame(
    {
        "CO2": [1.0, 2.0, 3.0],
    },
    geometry=[Point(0.75, 0.75), Point(0.25, 0.25), Point(1.2, 1)],
)
inv_with_pnt_sources.gdfs["liku"] = gpd.GeoDataFrame(
    {
        "CO2": [1.0, 2.0],
    },
    geometry=[Point(0.65, 0.75), Point(1.1, 0.8)],
)
inv_with_pnt_sources.gdfs["other"] = gpd.GeoDataFrame(
    {
        "AITS": [float(i + 1) for i in range(5)],
    },
    geometry=[
        Polygon(((0, 0), (0, 1), (1, 1), (1, 0))),
        Polygon(((0, 1), (0, 2), (1, 2), (1, 1))),
        Polygon(((1, 0), (1, 1), (2, 1), (2, 0))),
        Polygon(((1, 1), (1, 2), (2, 2), (2, 1))),
        Polygon(((2, 1), (2, 2), (3, 2), (3, 1))),
    ],
)
cropped = crop_with_shape(inv_with_pnt_sources, triangle, modify_grid=True)


def test_basic_crop():
    cropped = crop_with_shape(inv, triangle)


def test_with_modify_grid():
    cropped = crop_with_shape(inv, triangle, modify_grid=True)

    assert 4 not in cropped.gdf.index
    # Check an expected value
    assert cropped.gdf[("adf", "CH4")].iloc[0] == 3 / 8


def test_with_gdfs():
    cropped = crop_with_shape(inv_with_pnt_sources, triangle, modify_grid=True)

    assert len(cropped.gdfs["blek"]) == 2
    # Check  expected values
    assert cropped.gdfs["blek"]["CO2"].iloc[0] == 1 / 2  # At the boundary divided by 2
    assert cropped.gdfs["blek"]["CO2"].iloc[1] == 3.0

    # Check  point outside disappeared
    assert len(cropped.gdfs["liku"]) == 1
    # Make sure the index was reset
    assert 1 not in cropped.gdfs["liku"].index
    assert 0 in cropped.gdfs["liku"].index

    assert len(cropped.gdfs["other"]) == 3
    assert cropped.gdfs["other"]["AITS"].iloc[0] == 1 / 8
    assert cropped.gdfs["other"]["AITS"].iloc[1] == 3 / 4
    assert cropped.gdfs["other"]["AITS"].iloc[2] == 4 / 8


def test_with_modify_grid_and_cached():
    w_file = WEIGHTS_DIR / ".emiproc_test_with_modify_grid_and_cached"
    cropped = crop_with_shape(inv, triangle, weight_file=w_file, modify_grid=True)

    assert 4 not in cropped.gdf.index
    cropped = crop_with_shape(inv, triangle, weight_file=w_file, modify_grid=True)

    assert 4 not in cropped.gdf.index


def test_different_points_and_polygons_in_gdfs():
    inv = inv_only_one_gdfs

    cropped = crop_with_shape(inv, triangle)
    # outside shapes are removed

    assert len(cropped.gdfs["adf"]) == 5, "Did not crop expected number of shapes"
    expected_values = [1 / 2, 1.0, 1 / 8, 1 / 4, 1.0]  # At the boundary divided by 2
    # Check  expected values
    for i, val in enumerate(expected_values):
        assert cropped.gdfs["adf"]["CO2"].iloc[i] == val, f"Failed at {i}"


def test_no_emission():
    inv = inv_with_pnt_sources.copy()
    # Remove the only emission  column from the inv
    inv.gdfs["blek"].drop(columns=["CO2"], inplace=True)
    cropped = crop_with_shape(inv, triangle)

    assert "blek" not in cropped.gdfs
    assert "liku" in cropped.gdfs



"""plot an icon exported file.

This will generate a plot for each oem variable in the file.
"""
# %%
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from matplotlib.collections import PolyCollection
from matplotlib.colors import LogNorm

from emiproc import TESTS_DIR
from emiproc.plots.nclcmaps import cmap

# %% Load the file of icon
icon_file = TESTS_DIR / "export_icon" / "oem_gridded_emissions.nc"
out_folder = icon_file.parent / "plots"
out_folder.mkdir(exist_ok=True)
ds = xr.load_dataset(icon_file)

# %% choose the variables to plot
emiproc_generated_variables = [
    v for v in ds.variables if "created_by_emiproc" in ds[v].attrs
] + ["country_ids"]
# %% Loads the icon grid
n_cells = ds["cell"].size
corners = np.zeros((n_cells, 3, 2))
corners[:, :, 0] = ds["vlon"][ds["vertex_of_cell"] - 1].T
corners[:, :, 1] = ds["vlat"][ds["vertex_of_cell"] - 1].T
corners = np.rad2deg(corners)


# %%
# %matplotlib qt
matplotlib.style.use("default")
plt.ioff()
for var in emiproc_generated_variables:
    fig, ax = plt.subplots()

    da = ds[var]
    if not np.any(da):
        # No emissions at all
        continue
    
    min_val = da.min().values
    max_val = da.max().values

    if np.isnan(min_val) or np.isnan(max_val):
        print("skipping", var, "because of nan values in the data.")
        continue
    if min_val < 0:
        print("skipping", var, "because of negative values in the data.")
        continue
    poly_coll = PolyCollection(
        corners,
        cmap=cmap("WhViBlGrYeOrRe"),
        norm=LogNorm(),
        edgecolors="black",
        linewidth=0.04,
        antialiased=True,  # AA will help show the line when set to true
        alpha=0.6,
    )
    # Add the collection to the ax
    ax.add_collection(poly_coll)
    ax.set_ylim(np.min(corners[:, :, 1]), np.max(corners[:, :, 1]))
    ax.set_xlim(np.min(corners[:, :, 0]), np.max(corners[:, :, 0]))

    poly_coll.set_array(da)
    ax.set_title(da.attrs["long_name"])
    fig.colorbar(poly_coll)
    fig.savefig(out_folder / f"{var}.png")
    # fig.show()



"""Convert the mapluft inventory to gral.

For this exercise we need an additional package called pygg.
"""
# %%
from pathlib import Path
from emiproc.inventories.zurich import MapLuftZurich, ZURICH_SOURCES
from emiproc.exports.gral import export_to_gral
from emiproc.tests_utils import TEST_OUTPUTS_DIR
from emiproc.inventories.utils import crop_with_shape
from emiproc.inventories.utils import group_categories, validate_group
from emiproc.inventories.zurich.gral_groups import ZH_CO2_Groups
# pygg module for gram gral preprocessing
from pygg.grids import GralGrid
import numpy as np

# %%
file = Path("/store/empa/em05/mapluft/mapLuft_2020_v2021.gdb")

zh_inv = MapLuftZurich(file, ['CO2'], convert_lines_to_polygons=False)
zh_inv.emission_infos = ZURICH_SOURCES
zh_inv 
#%% Read the gral grid from a generated geb
grid = GralGrid.from_gral_rundir("/scratch/snx3000/lconstan/gramm_gral2/")

#%% Crop the invenotry over the gral grid 

zh_cropped = crop_with_shape(zh_inv, grid.get_bounding_polygon())

#%%

validate_group(ZH_CO2_Groups, zh_inv.categories)

# TODO: group should also group emission infos
zh_groupped = group_categories(zh_cropped, ZH_CO2_Groups, ignore_missing=False)

#%%
out_dir = TEST_OUTPUTS_DIR / 'test_gral_emissions'
out_dir.mkdir(exist_ok=True)
export_to_gral(
    zh_groupped,
    grid,
    out_dir,
    polygon_raster_size = 5
)
# %%



"""Inventories of emissions.

Contains the classes and functions to work with inventories of emissions.
"""

from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import dataclass
from os import PathLike
from typing import NewType, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr

from emiproc.grids import GeoPandasGrid, Grid
from emiproc.profiles import naming
from emiproc.profiles.temporal_profiles import (
    AnyTimeProfile,
    CompositeTemporalProfiles,
    TemporalProfile,
)
from emiproc.profiles.utils import check_valid_indexes
from emiproc.profiles.vertical_profiles import (
    VerticalProfile,
    VerticalProfiles,
    resample_vertical_profiles,
)

# Represent a substance that is emitted and can be present in a dataset.
Substance = NewType("Substance", str)
# Represent a category of emissions.
Category = NewType("Category", str)
# A colum from the gdf
CatSub = tuple[Category, Substance]

TemporalProfiles = Union[list[list[TemporalProfile]], CompositeTemporalProfiles]


@dataclass
class EmissionInfo:
    """Information about an emission category.

    This additional information is used for some models.
    It concerns only the :py:attr:`Inventory.gdfs` features.

    :param height: The height of the emission source (over the ground). [m]
    :param height_over_buildings: If True, the height is taken over buildings.
    :param width: The width of the emission source. [m]
    :param vertical_extension:
        The vertical extension (thickness) of the emission source. [m]
        This implies that the emission starts at height
        and ends at height + vertical_extension.
    :param temperature: The temperature of the emission source. [K]
    :param speed: The speed of the emission of the substances. [m/s]
    :param comment: A comment about the emission source.

    """

    # Height
    height: float = 0.0
    # weight the height is taken over buildings
    height_over_buildings: bool = True

    width: float = 0.5
    vertical_extension: float = 3.0
    temperature: float = 353.0
    speed: float = 5.0
    comment: str = ""


class Inventory:
    """Parent class for inventories.

    :param name: The name of the inventory. This is going to be used
        for adding metadata to the output files, and also for the reggridding
        weights files.
    :param year: The year of the inventory. (optional)
    :param grid: The grid on which the inventory is. Can be none if the invenotry
        is not defined on a grid. (only shapped emissions)
    :param substances: The :py:class:`Substance` present in this inventory.
    :param categories: List of the categories present in the inventory.

    :param emission_infos: Information about the emissions.
        Concerns only the :py:attr:`gdfs` features.
        This is optional, but mandatory for some models (ex. Gramm-Gral).

    :param gdf: The GeoPandas DataFrame that represent the whole inventory.
        The geometry column contains geometric objects for all the grid cells.
        The other columns should contain the emission value for the substances
        and the categories.

    :param gdfs: Some inventories are given on more than one grid.
        For example, :py:class:`MapLuftZurich` is given on a grid
        where every category has different shape file.
        In this case gdf must be set to None and gdfs will be
        a dictionnary mapping only the categories desired.

    :param v_profiles: A vertical profiles object.
    :param v_profiles_indexes: A :py:class:`xarray.DataArray` storing the information
        of which vertical profile belongs to which cell/category/substance.
        This allow to map each single emission value from the gdf to a specific
        profile.
        See :ref:`vertical_profiles` for more information.

    :param t_profiles_groups: A list  of temporal profiles groups.
        One temporal pattern can be defined by more than one temporal profile.
        (ex you can combine hour of day and day of week).
        The main list contains the different groups of temporal profiles.
        Each group is a list of :py:class:`TemporalProfile`.
    :param t_profiles_indexes: Same as :py:attr:`v_profiles_indexes`.
        For the temporal profiles, the indexes point to one of the groups.


    :param history: Stores all the operations that happened to this inventory.


    """

    name: str
    year: int | None = None

    grid: Grid | None
    substances: list[Substance]
    categories: list[Category]
    emission_infos: dict[Category, EmissionInfo]

    gdf: gpd.GeoDataFrame | None
    gdfs: dict[str, gpd.GeoDataFrame]
    geometry: gpd.GeoSeries

    v_profiles: VerticalProfiles | None = None
    v_profiles_indexes: xr.DataArray | None = None

    t_profiles_groups: TemporalProfiles | None = None
    t_profiles_indexes: xr.DataArray | None = None

    logger: logging.Logger
    history: list[str]

    def __init__(self) -> None:
        class_name = type(self).__name__
        if not hasattr(self, "name"):
            self.name = class_name
        self.history = [f"{self} created as type:'{class_name}'"]
        self.logger = logging.getLogger(f"emiproc.Inventory.{self.name}")

    def __repr__(self) -> str:
        return f"Inventory({self.name})"

    @property
    def emission_infos(self) -> dict[Category, EmissionInfo]:
        if hasattr(self, "_emission_infos"):
            return self._emission_infos
        else:
            raise ValueError(f"'emission_infos' were not set for {self}")

    @emission_infos.setter
    def emission_infos(self, emission_infos: dict[Category, EmissionInfo]):
        # Check that the cat is in the emission_infos
        missing_cats = set(self.categories) - set(emission_infos.keys())
        if missing_cats:
            raise ValueError(
                f"{missing_cats} are not in the emission_infos. "
                "Please add it to the emission_infos dict."
            )
        self._emission_infos = emission_infos

    @property
    def geometry(self) -> gpd.GeoSeries:
        return self.gdf.geometry

    @property
    def cell_areas(self) -> np.ndarray:
        """Area of the cells in m2 .

        These match the geometry from the gdf.
        """
        if not hasattr(self, "_cell_area"):
            # Compute with the grid
            if self.grid is None:
                raise ValueError(
                    "No grid set, cannot compute cell areas. "
                    f"implement or assign 'cell_areas' in {self.name}"
                )
            self._cell_area = np.array(self.grid.cell_areas)

        return self._cell_area

    @cell_areas.setter
    def cell_areas(self, cell_areas):
        if len(cell_areas) != len(self.gdf):
            raise ValueError(
                f"size does not match, got {len(cell_areas) }, expected {len(self.gdf)}"
            )

        self._cell_area = cell_areas

    @property
    def crs(self) -> int | None:
        if self.gdf is not None:
            return self.gdf.crs
        else:
            return self.gdfs[list(self.gdfs.keys())[0]].crs

    @property
    def categories(self) -> list[str]:
        return list(
            set(
                [
                    cat
                    for cat, _ in (self.gdf.columns if self.gdf is not None else [])
                    if not isinstance(self.gdf[(cat, _)].dtype, gpd.array.GeometryDtype)
                ]
            )
            | set(self.gdfs.keys())
        )

    @property
    def substances(self) -> list[Substance]:
        # Unique substances in the inventories
        subs = list(
            set(
                [
                    sub
                    for _, sub in (self.gdf.columns if self.gdf is not None else [])
                    if not isinstance(self.gdf[(_, sub)].dtype, gpd.array.GeometryDtype)
                ]
            )
            | set(sum([gdf.keys().to_list() for gdf in self.gdfs.values()], []))
        )
        if "geometry" in subs:
            subs.remove("geometry")
        return subs

    @property
    def total_emissions(self) -> pd.DataFrame:
        """Calculate the total emissions, returning a DataFrame.

        Simple accessor to the function
        :py:func:`~emiproc.inventories.utils.get_total_emissions`.
        """
        from emiproc.inventories.utils import get_total_emissions

        return pd.DataFrame(get_total_emissions(self)).T

    def copy(self, no_gdfs: bool = False, profiles: bool = True) -> Inventory:
        """Copy the inventory.

        :arg no_gdfs: Whether the gdfs should not be copied (main gdf and the gdfs).
        :arg profiles: Whether the profiles should be copied.
        """
        inv = Inventory()
        inv.__class__ = self.__class__
        inv.history = deepcopy(self.history)
        if hasattr(self, "grid"):
            inv.grid = self.grid

        if profiles and self.v_profiles is not None:
            inv.v_profiles = self.v_profiles.copy()
            inv.v_profiles_indexes = self.v_profiles_indexes.copy()
        if profiles and self.t_profiles_groups is not None:
            inv.t_profiles_groups = self.t_profiles_groups.copy()
            inv.t_profiles_indexes = self.t_profiles_indexes.copy()

        if no_gdfs or self.gdf is None:
            inv.gdf = None
        else:
            inv.gdf = self.gdf.copy(deep=True)

        if self.gdfs and not no_gdfs:
            inv.gdfs = {key: gdf.copy(deep=True) for key, gdf in self.gdfs.items()}
        else:
            inv.gdfs = {}

        # Copy the internal property
        if hasattr(self, "_emission_infos"):
            inv._emission_infos = deepcopy(self._emission_infos)

        inv.history.append(f"Copied from {type(self).__name__} to {inv}.")
        return inv

    @classmethod
    def from_gdf(
        cls,
        gdf: gpd.GeoDataFrame | None = None,
        name: str = "custom_from_gdf",
        gdfs: dict[str, gpd.GeoDataFrame] = {},
    ) -> Inventory:
        """The gdf must be a two level gdf with (category, substance)."""
        inv = Inventory()
        inv.name = name
        inv.gdf = gdf
        inv.gdfs = gdfs
        inv.grid = None if gdf is None else GeoPandasGrid(gdf)

        return inv

    @property
    def _gdf_columns(self) -> list[tuple[str, Substance]]:
        """All the columsn but not the geometric columns."""
        if self.gdf is None:
            return []
        return [
            col
            for col in self.gdf.columns
            if not isinstance(self.gdf[col].dtype, gpd.array.GeometryDtype)
        ]

    def to_crs(self, *args, **kwargs):
        """Same as geopandas.to_crs() but for inventories.

        Perform the conversion in place.
        """
        if self.gdf is not None:
            self.gdf.to_crs(*args, **kwargs, inplace=True)
        for gdf in self.gdfs.values():
            gdf.to_crs(*args, **kwargs, inplace=True)

    def set_crs(self, *args, **kwargs):
        """Same as geopandas.set_crs() but for inventories.

        Perform the conversion in place.
        """
        if self.gdf is not None:
            self.gdf.set_crs(*args, **kwargs, inplace=True)
        for gdf in self.gdfs.values():
            gdf.set_crs(*args, **kwargs, inplace=True)

    def add_gdf(self, category: Category, gdf: gpd.GeoDataFrame):
        """Add a gdf contaning emission sources to the inventory.

        This will add the category to the inventory and add the data to the
        inventory.gdfs dictionary.

        :arg category: The category to add.
        :arg gdf: The geodataframe containing the data for the category.
        """

        if category not in self.gdfs:
            self.gdfs[category] = gdf
            return

        # if the category is already present, we need to merge the data
        # this is a bit tricky, because we need to make sure that the
        # columns are the same

        # Set the values on missing columns to 0
        missing_columns = set(self.gdfs[category].columns) - set(gdf.columns)
        for col in missing_columns:
            gdf[col] = 0
        # Also set the substance to 0 in the gdfs if not there
        missing_columns = set(gdf.columns) - set(self.gdfs[category].columns)
        for col in missing_columns:
            self.gdfs[category][col] = 0

        # Now we can append
        self.gdfs[category] = pd.concat((self.gdfs[category], gdf), ignore_index=True)

    def set_profile(
        self,
        profile: VerticalProfile | list[TemporalProfile],
        category: str | None = None,
        substance: str | None = None,
    ) -> None:
        """Set a vertical or temporal profile to a specific category and/or substance.

        This happens in place.
        The profile is appened to the existing profiles.

        If only one of 'category' or 'substance' is specified, the profile is assigned
        to all the categories or substances on the non specified dimension.


        :arg profile: The profile to set.
            If Vertical profile, a vertical profile.
            If Temporal profile, a list of temporal profiles.
        :arg category: The category to set the profile to.
        :arg substance: The substance to set the profile to.
        """

        if isinstance(profile, VerticalProfile):
            indexes_array = self.v_profiles_indexes
            if self.v_profiles is None:
                # Set the profile for the first time
                self.v_profiles = VerticalProfiles(
                    ratios=np.array([profile.ratios]),
                    height=profile.height,
                )
            else:
                self.v_profiles = resample_vertical_profiles(
                    self.v_profiles, profile, specified_levels=self.v_profiles.height
                )
            profiles = self.v_profiles
        elif isinstance(profile, list):
            # Temporal profiles
            indexes_array = self.t_profiles_indexes
            if self.t_profiles_groups is None:
                self.t_profiles_groups = []
            self.t_profiles_groups.append(profile)

            profiles = self.t_profiles_groups
        else:
            raise ValueError(f"Unknown profile type {type(profile)}")

        if indexes_array is None:
            # Create it if it does not exist, axis is substance and category
            indexes_array = xr.DataArray(
                np.full(
                    (len(self.categories), len(self.substances)),
                    fill_value=-1,
                    dtype=int,
                ),
                dims=("category", "substance"),
                coords={"category": self.categories, "substance": self.substances},
            )

        # Add to the index the profile
        sel_dict = {}
        if category is not None:
            if "category" not in indexes_array.dims:
                # Exapnd the array over the categories of the inventory
                indexes_array = indexes_array.expand_dims({"category": self.categories})
            if category not in indexes_array.coords["category"].values:
                if category not in self.categories:
                    raise ValueError(
                        f"Category {category} is not in the inventory. "
                        "Please add it before setting the profile."
                    )
                # Otherwise add it to the coords and set the values to -1 (unassigned)
                # emtpy da with the new category being the only one in the
                new_cat_array = xr.DataArray(
                    -1,
                    dims=indexes_array.dims,
                    coords={
                        "category": [category],
                        **{
                            coord: indexes_array.coords[coord]
                            for coord in indexes_array.dims
                            if coord != "category"
                        },
                    },
                )
                indexes_array = xr.concat(
                    [indexes_array, new_cat_array], dim="category"
                )

            sel_dict["category"] = category
        if substance is not None:
            if "substance" not in indexes_array.dims:
                # Exapnd the array over the substnaces of the inventory
                indexes_array = indexes_array.expand_dims(
                    {"substance": self.substances}
                )
            sel_dict["substance"] = substance

        if sel_dict:
            # Set the index for the new profile
            profile_idx = len(profiles) - 1
            indexes_array.loc[sel_dict] = profile_idx
        else:
            self.logger.warning(
                "No category or substance specified for the profile."
                "The profile was not set."
            )
            return

        if isinstance(profile, VerticalProfile):
            self.v_profiles_indexes = indexes_array
        elif isinstance(profile, list):
            self.t_profiles_indexes = indexes_array

        self.history.append(f"Set profile {profile_idx} to {category}, {substance}.")

    def set_profiles(
        self,
        profiles: (
            VerticalProfiles | CompositeTemporalProfiles | list[list[AnyTimeProfile]]
        ),
        indexes: xr.DataArray,
    ):
        """Replace the profiles of the invenotry with the new profiles given.

        This checks that the indexes are correct and that the coords are matching.
        If they are not, gives a warning.

        If the profiles given are not valid (ex. not the same number of profiles
        as categories), raises an error.
        """

        indexes = indexes.copy()

        if isinstance(profiles, list):
            profiles = CompositeTemporalProfiles(profiles)

        # Check that the indexes are correct
        check_valid_indexes(indexes, profiles)

        # Check the coord if they match correctly what is given in the inventory
        for coord in naming.type_of_dim.keys():
            if coord not in indexes.dims:
                continue
            if coord == "cell":
                values_in_inv = list(range(len(self.grid)))
            elif coord == "category":
                values_in_inv = self.categories
            elif coord == "substance":
                values_in_inv = self.substances
            elif coord in ["country", "day_type", "time"]:
                # No check needed to be performed
                continue
            else:
                raise ValueError(f"Unknown coord {coord}")
            values_in_indexes = indexes.coords[coord].values
            values_not_in_inv = set(values_in_indexes) - set(values_in_inv)
            values_not_in_indexes = set(values_in_inv) - set(values_in_indexes)
            if values_not_in_inv:
                self.logger.warning(
                    f"{coord} {values_not_in_inv} from profiles are not in"
                    " the inventory."
                )
                indexes = indexes.drop_sel({coord: list(values_not_in_inv)})
            if values_not_in_indexes:
                self.logger.warning(
                    f"{coord} {values_not_in_indexes} from inventory are not in"
                    " the profiles."
                )

        # Assign to the inventory
        if isinstance(profiles, VerticalProfiles):
            self.v_profiles = profiles
            self.v_profiles_indexes = indexes
        elif isinstance(profiles, CompositeTemporalProfiles):
            self.t_profiles_groups = profiles
            self.t_profiles_indexes = indexes
        else:
            raise ValueError(f"Unknown profile type {type(profiles)}")


class EmiprocNetCDF(Inventory):
    """An output from emiproc.

    Useful if you need to process again an inventory.
    """

    def __init__(self, file: PathLike) -> None:
        super().__init__()


if __name__ == "__main__":
    test_inv = Inventory()



"""Some inventories that can be used for test purposes."""

import geopandas as gpd
from shapely.geometry import Point, Polygon

from emiproc.inventories import Inventory

from emiproc.tests_utils.test_grids import (
    basic_serie,
    basic_serie_2,
    basic_serie_of_size_2,
)


serie = basic_serie

inv = Inventory.from_gdf(
    gpd.GeoDataFrame(
        {
            ("adf", "CH4"): [i + 3 for i in range(len(serie))],
            ("adf", "CO2"): [i for i in range(len(serie))],
            ("liku", "CO2"): [i for i in range(len(serie))],
            ("test", "NH3"): [i + 1 for i in range(len(serie))],
        },
        geometry=serie,
    )
)

# An inventory with point sources
inv_with_pnt_sources = inv.copy()
inv_with_pnt_sources.gdfs["blek"] = gpd.GeoDataFrame(
    {
        "CO2": [1.0, 2.0, 3.0],
    },
    geometry=[Point(0.75, 0.75), Point(0.25, 0.25), Point(1.2, 1)],
)
inv_with_pnt_sources.gdfs["liku"] = gpd.GeoDataFrame(
    {
        "CO2": [1.0, 2.0],
    },
    geometry=[Point(0.65, 0.75), Point(1.1, 0.8)],
)
inv_with_pnt_sources.gdfs["other"] = gpd.GeoDataFrame(
    {
        "AITS": [1.0, 2.0],
    },
    geometry=[Point(0.65, 0.75), Point(1.1, 0.8)],
)

inv_only_one_gdfs = Inventory.from_gdf(
    gdfs={
        "adf": gpd.GeoDataFrame(
            {
                "CO2": [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            },
            geometry=[
                # corner point
                Point(0.75, 0.75),
                # Outside point
                Point(0.5, 0.4),
                # Inside point
                Point(1.2, 1),
                # 1/8 inside polygon
                Polygon(((0, 0), (0, 1), (1, 1), (1, 0))),
                # 1/4 inside polygon
                Polygon(((1, 0), (1, 1), (2, 1), (2, 0))),
                # outside polygon
                Polygon(((3, 3), (3, 4), (4, 4), (4, 3))),
                # fully inside polygon
                Polygon(((1, 0.5), (1.5, 0.5), (1.5, 1), (1, 1))),
            ],
        )
    }
)

inv_with_gdfs_bad_indexes = Inventory.from_gdf(
    gdfs={
        "adf": gpd.GeoDataFrame(
            {"CO2": [1.0, 2.0, 3.0]},
            geometry=[Point(0, 0), Point(1, 1), Point(2, 2)],
            index=[0, 1, 100000000],
        )
    }
)

inv_on_grid_serie2 = Inventory.from_gdf(
    gpd.GeoDataFrame(
        {
            ("adf", "CH4"): [i + 3 for i in range(len(basic_serie_2))],
            ("adf", "CO2"): [i for i in range(len(basic_serie_2))],
            ("liku", "CO2"): [i for i in range(len(basic_serie_2))],
            ("test", "NH3"): [i + 1 for i in range(len(basic_serie_2))],
        },
        geometry=basic_serie_2,
    )
)

inv_on_grid_serie2_bis = Inventory.from_gdf(
    gpd.GeoDataFrame(
        {
            ("adf", "CH4"): [i + 3 for i in range(len(basic_serie_of_size_2))],
            ("adf", "CO2"): [i for i in range(len(basic_serie_of_size_2))],
            ("liku", "CO2"): [i for i in range(len(basic_serie_of_size_2))],
            ("test", "NH3"): [i + 1 for i in range(len(basic_serie_of_size_2))],
        },
        geometry=basic_serie_of_size_2,
    )
)



"""Convert the TNO inventory to a netcdf file on a regular raster grid."""

#%%
from pathlib import Path

from emiproc.inventories.tno import TNO_Inventory

from emiproc.utilities import Units
from emiproc.grids import RegularGrid
from emiproc.inventories.utils import group_categories

from emiproc.exports.netcdf import nc_cf_attributes
from emiproc.exports.rasters import export_raster_netcdf
from emiproc.exports.profiles import export_inventory_profiles

#%%
# path to input inventory
input_path = Path("/scratch/snx3000/lconstan/TNO_GHGco_v4_1_highres_year2018.nc")
# output path and filename
output_dir = input_path.parent
output_path = output_dir / f"{input_path.stem}_with_emissions.nc"

#%%
inv = TNO_Inventory(
    input_path,
    # Take only CH4
    substances=["CH4", 'CO2'],
)


# %%  mapping from input to output species
mapping = {
    "energy": ["A"],
    "industrial": ["B"],
    "residential": ["C"],
    "natural_gas": ["D", "E"],
    "transport": ["F1", "F2", "F3", "F4", "G", "H", "I"],
    "waste": ["J"],
    "agriculture": ["K", "L"],
}


groupped_tno = group_categories(inv, mapping)


# %% Regular grid domain, If you want to use the TNO grid, use output_grid = inv.grid


output_grid = RegularGrid(
    xmin=4.97, ymin=45.4875, xmax=11.21, ymax=48.5, nx=312, ny=208
)

#%%


nc_metadata = nc_cf_attributes(
    author="Stephan Henne",
    contact="stephan.henne@empa.ch",
    title="Gridded annual emissions",
    source=input_path.name,
    institution="Empa Duebendorf, Switzerland",
)

# Export to a netcdf raster on the output grid, remapping will be performed
export_raster_netcdf(
    groupped_tno, output_path, output_grid, nc_metadata, unit=Units.KG_PER_M2_PER_S
)


# %% Export the temporal profiles of TNO
export_inventory_profiles(
    inv=groupped_tno,
    output_dir=output_dir,
    grid=output_grid,
    netcdf_attributes=nc_metadata,
)
# %%



"""Test for the vprm module.

Tests are there moslty to check that things run problerly and that the inputs/
outputs expected are there. The actual values are not checked.
"""
import pytest
import pandas as pd

import numpy as np
from emiproc.profiles.vprm import calculate_vprm_emissions, calculate_vegetation_indices


@pytest.fixture
def sample_data():
    df = pd.DataFrame({
        ('T', 'global'): [25, 30, 28, 27],
        ('T', 'urban'): [28, 32, 30, 29],
        ('vegetation_type_1', 'lswi'): [0.5, 0.6, 0.7, 0.8],
        ('vegetation_type_1', 'evi'): [0.2, 0.3, 0.4, 0.5],
        ('vegetation_type_2', 'lswi'): [0.3, 0.4, 0.5, 0.6],
        ('vegetation_type_2', 'evi'): [0.1, 0.2, 0.3, 0.4],
        ('RAD', ''): [100, 200, 150, 180]
    })
    
    df_vprm = pd.DataFrame({
        'alpha': [0.1, 0.2],
        'beta': [0.5, 0.6],
        'lambda': [0.8, 0.9],
        'Tmin': [20, 22],
        'Topt': [25, 27],
        'Tmax': [30, 32],
        'Tlow': [18, 20],
        'PAR0': [50, 60],
        'is_urban': [False, True]
    }, index=['vegetation_type_1', 'vegetation_type_2'])
    
    return df, df_vprm

def test_calculate_vprm_emissions(sample_data):
    df, df_vprm = sample_data
    result = calculate_vprm_emissions(df, df_vprm)
    
    assert ('PAR', '') in result.columns
    assert ('vegetation_type_1', 'Tscale') in result.columns
    assert ('vegetation_type_2', 'Tscale') in result.columns
    assert ('vegetation_type_1', 'Wscale') in result.columns
    assert ('vegetation_type_2', 'Wscale') in result.columns
    assert ('vegetation_type_1', 'Pscale') in result.columns
    assert ('vegetation_type_2', 'Pscale') in result.columns
    assert ('vegetation_type_1', 'resp') in result.columns
    assert ('vegetation_type_2', 'resp') in result.columns
    assert ('vegetation_type_1', 'gee') in result.columns
    assert ('vegetation_type_2', 'gee') in result.columns
    assert ('vegetation_type_1', 'nee') in result.columns
    assert ('vegetation_type_2', 'nee') in result.columns




def test_calculate_vegetation_indices():
    nir = np.array([0.8, 0.9, 0.7, 0.6])
    swir = np.array([0.4, 0.5, 0.3, 0.2])
    red = np.array([0.6, 0.7, 0.5, 0.4])
    blue = np.array([0.2, 0.3, 0.1, 0.0])


    evi, lswi, ndvi = calculate_vegetation_indices(nir, swir, red, blue)

    # Check that the arrays have the right shape
    assert evi.shape == (4,)
    assert lswi.shape == (4,)
    assert ndvi.shape == (4,)



"""Test some basic features of the inventory profiles."""
import xarray as xr
from pathlib import Path
import numpy as np
import pytest
import geopandas as gpd
from shapely.geometry import Polygon

from emiproc.inventories import Inventory
from emiproc.inventories.utils import group_categories
from emiproc.profiles.vertical_profiles import (
    VerticalProfile,
    VerticalProfiles,
    check_valid_vertical_profile,
    get_mid_heights,
    resample_vertical_profiles,
    get_weights_profiles_interpolation,
    get_delta_h,
)
from emiproc.profiles.operators import (
    group_profiles_indexes,
    weighted_combination,
    combine_profiles,
    get_weights_of_gdf_profiles,
)

from emiproc.tests_utils import vertical_profiles


def test_test_profiles():
    for p in vertical_profiles.list_of_three:
        check_valid_vertical_profile(p)
    check_valid_vertical_profile(vertical_profiles.VerticalProfiles_instance)


def test_weights_interpolation():
    from_h = np.array([5, 10, 40, 100, 120])
    levels = np.array([2, 8, 10, 50, 80])
    w = get_weights_profiles_interpolation(from_h, levels)

    assert w[0, 0] == 2 / 5
    assert w[-1, -1] == 1.0
    assert w[1, 0] == 3 / 5


def test_rescale():
    # We just test it runs
    new_profiles = resample_vertical_profiles(
        *vertical_profiles.list_of_three,
        vertical_profiles.VerticalProfiles_instance,
        specified_levels=[0, 10, 20, 40, 50, 60],
    )

    # Visual check
    # start_from_0 = lambda arr: np.hstack([0, arr])
    # start_from_1 = lambda arr: np.hstack([1, arr])
    # for i, profile in enumerate(list_of_three):
    #     plt.figure()
    #     plot_profile = lambda h, r, **kwargs: plt.step(
    #         start_from_0(h),
    #         # np.cumsum(start_from_0(r)),
    #         start_from_0(r) / start_from_1(get_delta_h(h)),
    #         where="pre",
    #         **kwargs
    #     )
    #     plot_profile(new_profiles.height, new_profiles.ratios[i])
    #     plot_profile(profile.height, profile.ratios, linestyle='--')
    #     plt.show()


def test_mid_heights():
    assert np.allclose(
        get_mid_heights(np.array([10, 20, 40, 50])), np.array([5, 15, 30, 45])
    )


# test addition
def test_addition_vertical_profiles():
    new_profiles = (
        vertical_profiles.VerticalProfiles_instance
        + vertical_profiles.VerticalProfiles_instance
    )
    assert (
        new_profiles.n_profiles
        == 2 * vertical_profiles.VerticalProfiles_instance.n_profiles
    )

    # Check we can also do +=
    new_profiles += vertical_profiles.VerticalProfiles_instance
    assert (
        new_profiles.n_profiles
        == 3 * vertical_profiles.VerticalProfiles_instance.n_profiles
    )


#
def test_weighted_combination():
    weights = np.array([1, 2, 3])
    new_profile = weighted_combination(
        vertical_profiles.VerticalProfiles_instance, weights=weights
    )
    check_valid_vertical_profile(new_profile)
    new_total_emissions = np.sum(weights) * new_profile.ratios
    previous_total_emissions = weights.dot(
        vertical_profiles.VerticalProfiles_instance.ratios
    )
    # The combination should give the same as summing the emissions one by one
    assert np.allclose(
        new_total_emissions, previous_total_emissions
    ), f"{new_total_emissions},{previous_total_emissions}"


#
def test_invalid_vertical_profiles():
    # invalid profiles should raise some errors, let's check the main ones
    with pytest.raises(AssertionError):
        # Height 0
        check_valid_vertical_profile(VerticalProfile(np.array([1]), np.array([0])))
    with pytest.raises(AssertionError):
        # Sum of ratio != 1
        check_valid_vertical_profile(VerticalProfile(np.array([2]), np.array([10])))
    with pytest.raises(AssertionError):
        # non increasing heights
        check_valid_vertical_profile(
            VerticalProfile(np.array([0.5, 0.0, 0.5]), np.array([10, 40, 30]))
        )
    with pytest.raises(AssertionError):
        # ration < 0
        check_valid_vertical_profile(
            VerticalProfile(np.array([-0.5, 1.5]), np.array([10, 20]))
        )
    with pytest.raises(AssertionError):
        # Different length
        check_valid_vertical_profile(
            VerticalProfile(np.array([0.5, 0.5]), np.array([10, 20, 30]))
        )
    with pytest.raises(AssertionError):
        # nan value
        check_valid_vertical_profile(
            VerticalProfile(np.array([0.5, 0.5, np.nan]), np.array([10, 20, 30]))
        )


# select a subset

categories = ["test_cat", "test_cat2"]
selected_profile_indexes = vertical_profiles.corresponding_vertical_profiles.sel(
    {"category": categories}
)

# get the weights associatied to each profile


weights = get_weights_of_gdf_profiles(vertical_profiles.gdf, selected_profile_indexes)

# Make an average over the category taking care of putting the weights on the profiles where they should go


def test_combination_over_dimensions():
    new_profiles, new_indexes = combine_profiles(
        vertical_profiles.VerticalProfiles_instance,
        selected_profile_indexes,
        "cell",
        weights,
    )

    # Missing data
    assert all(new_indexes.sel(dict(substance="NH3")) == -1)
    # The missing data must have been merged (not all were missing)
    assert not any(new_indexes.sel(dict(substance="CH4")) == -1)
    # Access the profiles we want to compare (orginal of cat 0 and 1 for CH4 )
    assert np.allclose(
        new_profiles.ratios[
            new_indexes.sel(dict(substance="CH4", category="test_cat"))
        ],
        np.array([0.0, 0.3, 6.8 / 14, 3.0 / 14]),
    )
    assert np.allclose(
        new_profiles.ratios[
            new_indexes.sel(dict(substance="CH4", category="test_cat2"))
        ],
        np.array([0.08, 0.22, 0.7, 0.0]),
    )

    check_valid_vertical_profile(new_profiles)
    new_profiles, new_indexes = combine_profiles(
        vertical_profiles.VerticalProfiles_instance,
        selected_profile_indexes,
        "category",
        weights,
    )
    # Missing data
    assert all(new_indexes.sel(dict(substance="NH3")) == -1)
    assert new_indexes.sel(dict(cell=1, substance="CH4")) == -1
    # Access the profiles we want to compare (orginal of cells 0, 2, 3 for CH4 )
    indexes_of_CH4 = new_indexes.sel(dict(substance="CH4"))
    assert np.all(
        new_profiles.ratios[indexes_of_CH4[0]]
        == vertical_profiles.VerticalProfiles_instance.ratios[0]
    )
    assert np.all(
        new_profiles.ratios[indexes_of_CH4[2]]
        == vertical_profiles.VerticalProfiles_instance.ratios[0]
    )
    assert np.allclose(
        new_profiles.ratios[indexes_of_CH4[3]], np.array([0.04, 0.26, 0.4, 0.3])
    )

    check_valid_vertical_profile(new_profiles)
    new_profiles, new_indexes = combine_profiles(
        vertical_profiles.VerticalProfiles_instance,
        selected_profile_indexes,
        "substance",
        weights,
    )
    check_valid_vertical_profile(new_profiles)

def test_group_profiles_categories():

    weights = get_weights_of_gdf_profiles(vertical_profiles.inv.gdf, vertical_profiles.inv.v_profiles_indexes)
    new_profiles,new_indices = group_profiles_indexes(
        profiles = vertical_profiles.inv.v_profiles,
        profiles_indexes = vertical_profiles.inv.v_profiles_indexes,
        indexes_weights=weights,
        categories_group=vertical_profiles.inv_groups_dict,
    )
    
def test_group_profiles_substances():

    weights = get_weights_of_gdf_profiles(vertical_profiles.inv.gdf, vertical_profiles.inv.v_profiles_indexes)
    new_profiles,new_indices = group_profiles_indexes(
        profiles = vertical_profiles.inv.v_profiles,
        profiles_indexes = vertical_profiles.inv.v_profiles_indexes,
        indexes_weights=weights,
        categories_group=vertical_profiles.inv_groups_subs_dict,
        groupping_dimension="substance"
    )
    
    

def test_can_group_profiles_in_inv():
    group_categories(
        vertical_profiles.inv, vertical_profiles.inv_groups_dict
    )



def test_point_sources_are_just_appened():
    g_inv = group_categories(vertical_profiles.inv, {"new_cat": ["test_cat", "test_cat2", "test_cat3"]})

    assert len(g_inv.gdfs["new_cat"]) == 8
    # Check that the data was correctly appended
    assert np.all(g_inv.gdfs["new_cat"]["CO2"] == [1, 2, 3, 1.2, 2.7, 8, 1, 2])
    assert np.all(g_inv.gdfs["new_cat"]["CH4"] == [0, 0, 0, 4, 2, 8, 0, 0])
    assert np.all(g_inv.gdfs["new_cat"]["__v_profile__"] == [-1, 2, 1, 1, 2, -1, -1, -1])


#  Create hourly profiles
def check_valid_profiles_indexes_array(inv: Inventory, index_array: xr.DataArray):
    for cat, sub in inv._gdf_columns:
        print(cat)
        indexes = {}
        # get the values from the profile
        if "substance" in index_array.dims:
            if sub not in index_array["substance"]:
                raise ValueError(f"Missing profile for substance {sub}")
            indexes["substance"] = sub
        if "category" in index_array.dims:
            if cat not in index_array["category"]:
                raise ValueError(f"Missing profile for category {cat}")
            indexes["category"] = sub

        total_emission_da = xr.DataArray(
            inv.gdf[(cat, sub)], {"cells": np.arange(len(inv.gdf))}
        )

        profiles_this_catsub = index_array.loc[indexes]

        gridded_data = total_emission_da * profiles_this_catsub


if __name__ == "__main__":
    pytest.main([__file__])
    #test_group_profiles_substances()



# Cateogies taken from the Bericht CO2_GRAMM_GRAL_Bericht_UGZ_20230207
# Any change is written there as comment
# Categories not having any CO2 are then not included
# There are 4 categories that have been ignored
# 'c5801_BrandFeuerschaeden_Emissionen_Kanton'
# 'c3418_Vergaerwerk_Emissionen_Kanton'
# 'c5601_Feuerwerke_Emissionen_Kanton'
# 'c5701_Tabakwaren_Emissionen_Kanton'
ZH_CO2_Groups = {
    # Shipping
    "Schiffahrt": [
        "c1101_Linienschiffe_Emissionen_Kanton",
        "c1102_PrivaterBootsverkehr_Emissionen_Kanton",
    ],
    # Industry
    "Industrie": [
        "c3201_Notstromanlagen_Emissionen_Kanton",
        "c3301_Prozessenergie_Emissionen_Kanton",
        # "c3401_Metallreinigung_Emissionen_Kanton", # not in bericht
        # "c3402_Holzbearbeitung_Emissionen_Kanton", # not in bericht
        # "c3403_Malereien_Emissionen_Kanton", # not in bericht
        # "c3404_Textilreinigung_Emissionen_Kanton", # not in bericht
        # "c3405_Karosserien_Emissionen_Kanton", # not in bericht
        # "c3406_Raeuchereien_Emissionen_Kanton", # not in bericht
        # "c3407_Roestereien_Emissionen_Kanton", # not in bericht
        # "c3408_Druckereien_Emissionen_Kanton", # not in bericht
        # "c3409_Laboratorien_Emissionen_Kanton", # not in bericht
        "c3410_Bierbrauereien_Emissionen_Kanton",
        "c3411_Brotproduktion_Emissionen_Kanton",
        # "c3412_MedizinischePraxen_Emissionen_Kanton", # not in bericht
        # "c3413_Gesundheitswesen_Emissionen_Kanton", # not in bericht
        "c3414_Krematorium_Emissionen_Kanton",
        "c3416_Tankstellen_Emissionen_Kanton",
    ],
    # Fossil heating
    "FeuerungenFossil": [
        "c2101_Oelheizungen_Emissionen_Kanton",
        "c2102_Gasheizungen_Emissionen_Kanton",
        "c2105_Warmwassererzeuger_Emissionen_Kanton",
        "c2201_BHKW_Emissionen_Kanton",
    ],
    # Heating
    "Feuerungen": [
        "c2103_HolzheizungenLokalisiert_Emissionen_Kanton",
        "c2104_HolzheizungenDispers_Emissionen_Kanton",
        "c2401_Klaerschlammverwertung_Emissionen_Kanton",
    ],
    # Heting Power Plants
    "KehrichtheizkraftwerkeKHKW": [
        "c2301_KHKWKehricht_Emissionen_Kanton",
        "c2302_KHKWErdgas_Emissionen_Kanton",
        "c2303_KHKWHeizoel_Emissionen_Kanton",
    ],
    # # Solvents and product use
    # "GNFR_E": [
    #     "c3417_LoesemittelIG_Emissionen_Kanton",
    #     "c5101_LoesemittelHH_Emissionen_Kanton",
    # ],
    # Road transport
    "Strassenverkehr": [
        "c1301_Personenwagen_Emissionen_Kanton",
        "c1303_Motorraeder_Emissionen_Kanton",
        "c1306_StartStopTankatmung_Emissionen_Kanton",
        "c1307_Lieferwagen_Emissionen_Kanton",
        "c1308_Reisebusse_Emissionen_Kanton",
    ],
    # Heavy transport
    "Schwerverkehr": [
        "c1302_Lastwagen_Emissionen_Kanton",
    ],
    # Public transport
    "OffentlicherVerkehr ": [
        "c1304_Linienbusse_Emissionen_Kanton",
        # "c1305_Trolleybusse_Emissionen_Kanton", # This is assumed from the bericht
    ],
    # Offroad mobility
    # "GNFR_I": [
    #    "c1201_BahnPersonenverkehr_Emissionen_Kanton",
    #    "c1202_BahnGueterverkehr_Emissionen_Kanton",
    #    "c1203_Tramverkehr_Emissionen_Kanton",
    #    "c1204_Kleinbahnen_Emissionen_Kanton",
    #    # c31xx are construction stuff
    #    "c3102_Bitumen_Emissionen_Kanton",
    #    "c3103_FarbenBaustelle_Emissionen_Kanton",
    #    "c3105_Strassenbelag_Emissionen_Kanton",
    # ],
    "FahrzeugeMaschinen": [
        "c3101_MaschinenHochbau_Emissionen_Kanton",
        "c3104_MaschinenTiefbau_Emissionen_Kanton",
        "c3419_IndustrielleFZ_Emissionen_Kanton",
        "c4101_ForstwirtschaftlicheFZ_Emissionen_Kanton",
        "c4201_LandwirtschaftlicheFZ_Emissionen_Kanton",
    ],
    # Waste
    "Umschwung": [
        "c5201_Gruenabfallverbrennung_Emissionen_Kanton",
        "c5301_HolzoefenKleingarten_Emissionen_Kanton",
        "c5401_AbfallverbrennungHaus_Emissionen_Kanton",
        # "c3418_Vergaerwerk_Emissionen_Kanton",
    ],
    # # AgriLivestock
    # "GNFR_K": [
    #     "c4401_Nutztierhaltung_Emissionen_Kanton",
    # ],
    # # AgriOther
    # "GNFR_L": [
    #     "c4301_Nutzflaechen_Emissionen_Kanton",
    # ],
    # # Others
    # "GNFR_R": [
    #     "c5501_HausZooZirkustiere_Emissionen_Kanton",
    #     "c5601_Feuerwerke_Emissionen_Kanton",
    #     "c5701_Tabakwaren_Emissionen_Kanton",
    #     "c5801_BrandFeuerschaeden_Emissionen_Kanton",
    #     "c6101_Waelder_Emissionen_Kanton",
    #     "c6201_Grasflaechen_Emissionen_Kanton",
    #     "c6301_Gewaesser_Emissionen_Kanton",
    #     "c6401_Blitze_Emissionen_Kanton",
    # ],
}

# Swiss invenotry is also used to be remapped on the the boundaries around zuirch
# Note missing co2 categories: many have been removeved becaue not in the zh area
# eipwp # Industrial point sources 
# eipzm # Zement Werke
# evsfa # Shiffarts lilinen should be verkeher
# evsrh # Rehin shiffarts, not in zh but still 
# eipis # pisten fahrzeuge 
# evsra # Schienenverkehr Rangieren 
# elfer # landwirtschaftliche Nutzflächen
# ehhab # Hausalte Brande
# ehhaf # Haushalte andere Feuerwerk etc
CH_2_GNFR = {
    # Road transport
    "Verkehr": [
        # "evstr_ch4",
        # "evstr_co",
        # "evstr_co2",
        # "evstr_n2o",
        # "evstr_nh3",
        # "evstr_nmvoc",
        # "evstr_nox",
        # "evstr_so2",
        "evstr",
        "evzon",
        "evsee",
    ],
    "FeuerungenFossil": [
        "ehfoe",
        "ehgws",
        "ehare",
    ],
    "Feuerungen": [
        "ehfho",
    ],
    "IndustrieUndRest": [
        "ehhan",
        "ehmgh",
        "eibau",
        "eifrz",
        "eilmi",  # Lösungsmittel Industrie
        "eilpf",
        "eiprd",
        "eipkv",  # Punktquellen KVA (Kehrichtverbrennungsanlagen ) == Waste incinerators
        "evsch",
        "evfgva",  # GVA airport, should not be in the area but put there as zh is there
        "evfzhr",  # ZH airport
        "elfwm",
        "ellwm",
        "eivgn",
        "eipro",
    ],



    # # Industry
    # "GNFR_B": [
    #     "eipwp",  # this is the weitere punktquelle (additional point sources)
    #     "eipzm",
    # ],
    # # Other stationary combustion (services, residential, agriculture)
    # "GNFR_C": [
    #     "eipdh",
    #     "elfeu",
    # ],
    # # Fugitives
    # "GNFR_D": [
    #     "eilgk",
    #     "evklm",
    #     "evtrk",
    # ],
    # # Solvents and product use
    # "GNFR_E": [
    #     "ehlmk",  # Lösungsmittel Konsumprodukte
    # ],
    # # Shipping
    # "GNFR_G": [
    #     "evsfa",
    #     "evsrh",
    # ],
    # # Aviation
    # "GNFR_H": [],
    # # Offroad mobility
    # "GNFR_I": [
    #     "eipis",
    #     "evsra",
    # ],
    # # Waste
    # "GNFR_J": [
    #     "eidep",
    #     "eikla",
    #     "eikmp",
    #     "elabf",
    #     "elver",
    # ],
    # # AgriLivestock
    # "GNFR_K": [
    #     "elapp",
    #     "elsto",
    # ],
    # #  AgriOther
    # "GNFR_L": [
    #     "elfer",
    # ],
    # # Others
    # "GNFR_R": [
    #     "ehhab",
    #     "ehhaf",
    #     "enwal",
    # ],
}



"""Utilities for testing emiproc."""
import emiproc
from pathlib import Path

TEST_DIR = Path(*emiproc.__path__) / ".." / "tests"
TEST_DIR.mkdir(exist_ok=True)

TESTFILES_DIR = emiproc.TESTS_DIR

WEIGHTS_DIR = TEST_DIR / ".weights"
WEIGHTS_DIR.mkdir(exist_ok=True)

TEST_OUTPUTS_DIR = TEST_DIR / ".outputs"
TEST_OUTPUTS_DIR.mkdir(exist_ok=True)


if __name__ == "__main__":
    print(WEIGHTS_DIR)
    print(WEIGHTS_DIR.is_dir())



"""Emiproc Grids.

Classes handling different grids, namely the simulation grids and
grids used in different emissions inventories.
"""

from __future__ import annotations

import math
import warnings
from functools import cache, cached_property
from typing import Iterable

import geopandas as gpd
import numpy as np
import pyproj
import xarray as xr
from netCDF4 import Dataset
from shapely.geometry import LineString, MultiPolygon, Point, Polygon, box
from shapely.ops import split
from shapely.creation import polygons

WGS84 = 4326
WGS84_PROJECTED = 3857
LV95 = 2056  # EPSG:2056, swiss CRS, unit: meters
WGS84_NSIDC = 6933  # Unit: meters

# Radius of the earth
R_EARTH = 6371000  # m


# Type alias
# minx, miny, maxx, maxy
BoundingBox = tuple[float, float, float, float]


class Grid:
    """Abstract base class for a grid.
    Derive your own grid implementation from this and make sure to provide
    an appropriate implementation of the required methods.
    As an example you can look at TNOGrid.
    """

    nx: int
    ny: int

    # The crs value as an integer
    crs: int | str
    gdf: gpd.GeoDataFrame

    def __init__(self, name: str, crs: int | str = WGS84):
        """
        Parameters
        ----------
        name : Name of the grid.
        crs : The coordinate reference system of the grid.
        """
        self.name = name
        self.crs = crs

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name})"

    @property
    def gdf(self) -> gpd.GeoDataFrame:
        """Return a geopandas dataframe containing the grid."""
        if not hasattr(self, "_gdf"):
            self._gdf = gpd.GeoDataFrame(
                geometry=self.cells_as_polylist,
                crs=self.crs,
            )
        return self._gdf

    @gdf.setter
    def gdf(self, value: gpd.GeoDataFrame):
        warnings.warn(
            "deprectated to set the gdf of a grid. It is now automatically generated."
        )
        self._gdf = value

    def cell_corners(self, i, j):
        """Return the corners of the cell with indices (i,j).

        The points are ordered clockwise, starting in the top
        left:

        4.   1.
        ^    v
        3. < 2.

        Returns a tuple of arrays with shape (4,). The first
        tuple element are the x-coordinates of the corners,
        the second are the y-coordinates.

        The coordinates are in the projection of the grid, so
        to work with them you might have to transform them to
        the desired projection. For example, to be sure you're
        working with regular (lon, lat) coordinates:

        >>> corners = ccrs.PlateCarree().transform_points(
        ...    grid.get_projection(),
        ...    *grid.cell_corners(i,j)
        ...)

        The clunky return type is necessary because the corners
        are transformed after by cartopy.crs.CRS.transform_points.

        Parameters
        ----------
        i : int
        j : int

        Returns
        -------
        tuple(np.array(shape=(4,), dtype=float),
              np.array(shape=(4,), dtype=float))
            Arrays containing the x and y coordinates of the corners

        """
        raise NotImplementedError("Method not implemented")

    @cached_property
    def cells_as_polylist(self) -> list[Polygon]:
        """Return all the cells as a list of polygons."""
        return [
            Polygon(zip(*self.cell_corners(i, j)))
            for i in range(self.nx)
            for j in range(self.ny)
        ]

    @cached_property
    def shape(self) -> tuple[int, int]:
        return (self.nx, self.ny)

    @cached_property
    def cell_areas(self) -> Iterable[float]:
        """Return an array containing the area of each cell in m2."""
        return (
            gpd.GeoSeries(self.cells_as_polylist, crs=self.crs)
            # Convert to WGS84 to get the area in m^2
            .to_crs(epsg=WGS84_NSIDC).area
        )

    def __len__(self):
        """Return the number of cells in the grid."""
        return self.nx * self.ny


class RegularGrid(Grid):
    """Regular grid a grids with squared cells.

    This allows for some capabilities that are not available for
    irregular grids (rasterization, image like plotting).

    The grid can be defined in multiple ways.
    All the way need the reference (xmin, ymin).
    Then you need 2 of the three following:

    * xmax, ymax to define the bounding box
    * nx, ny to define the number of cells in each direction
    * dx, dy to define the size of the cells in each direction

    Leave the unused parameters as None.
    """

    # The centers of the cells (lon =x, lat = y)
    lon_range: np.ndarray
    lat_range: np.ndarray

    # The edges of the cells
    lat_bounds: np.ndarray
    lon_bounds: np.ndarray

    xmin: float
    xmax: float
    ymin: float
    ymax: float

    dx: float
    dy: float

    def __init__(
        self,
        xmin: float,
        ymin: float,
        xmax: float | None = None,
        ymax: float | None = None,
        nx: int | None = None,
        ny: int | None = None,
        dx: float | None = None,
        dy: float | None = None,
        name: str | None = None,
        crs: int | str = WGS84,
    ):
        self.xmin, self.ymin = xmin, ymin

        # Check if they did not specify all the optional parameters
        if all((p is not None for p in [xmax, ymax, nx, ny, dx, dy])):
            raise ValueError(
                "Specified too many parameters. "
                "Specify only 2 of the following: "
                "(xmax, ymax), (nx, ny), (dx, dy)"
            )

        if dx is None and dy is None and xmax is None and ymax is None:
            raise ValueError(
                "Cannot create grid with only nx and ny. "
                "Specify at least dx, dy or xmax, ymax."
            )

        # Calclate the number of cells if not specified
        if nx is None and ny is None:
            if dx is None or dy is None:
                raise ValueError(
                    "Either nx and ny or dx and dy must be specified. "
                    f"Received: {nx=}, {ny=}, {dx=}, {dy=}"
                )
            if xmax is None or ymax is None:
                raise ValueError(
                    "When using only dx dy, xmax and ymax must be specified. "
                    f"Received: {xmax=}, {ymax=}"
                )
            # Guess the nx and ny values, override the max
            nx = math.ceil((xmax - xmin) / dx)
            ny = math.ceil((ymax - ymin) / dy)

        if xmax is None and ymax is None:
            xmax = xmin + nx * dx
            ymax = ymin + ny * dy
        self.xmax, self.ymax = xmax, ymax

        # Calculate all grid parameters
        self.nx, self.ny = nx, ny
        self.dx, self.dy = (xmax - xmin) / nx, (ymax - ymin) / ny

        self.lon_range = np.arange(xmin, xmax, self.dx) + self.dx / 2
        self.lat_range = np.arange(ymin, ymax, self.dy) + self.dy / 2

        self.lon_bounds = np.concatenate(
            [self.lon_range - self.dx / 2, [self.lon_range[-1] + self.dx / 2]]
        )
        self.lat_bounds = np.concatenate(
            [self.lat_range - self.dy / 2, [self.lat_range[-1] + self.dy / 2]]
        )

        assert len(self.lon_range) == nx, f"{len(self.lon_range)=} != {nx=}"
        assert len(self.lat_range) == ny, f"{len(self.lat_range)=} != {ny=}"
        assert len(self.lon_bounds) == nx + 1
        assert len(self.lat_bounds) == ny + 1

        if name is None:
            name = f"x({xmin},{xmax})_y({ymin},{ymax})_nx({nx})_ny({ny})"

        super().__init__(name, crs)

    @cached_property
    def cells_as_polylist(self) -> list[Polygon]:

        x_coords, y_coords = np.meshgrid(
            self.lon_range - self.dx / 2.0, self.lat_range - self.dy / 2.0
        )
        # Reshape to 1D (order set for backward compatibility)
        x_coords = x_coords.flatten(order="F")
        y_coords = y_coords.flatten(order="F")
        dx = float(self.dx)
        dy = float(self.dy)
        coords = np.array(
            [
                [x, y]
                for x, y in zip(
                    [x_coords, x_coords, x_coords + dx, x_coords + dx],
                    [y_coords, y_coords + dy, y_coords + dy, y_coords],
                )
            ]
        )
        coords = np.rollaxis(coords, -1, 0)
        return polygons(coords)

    def cell_corners(self, i, j):
        """Return the corners of the cell with indices (i,j).

        The points are ordered clockwise, starting in the top
        right:

        """
        x = self.xmin + i * self.dx
        y = self.ymin + j * self.dy

        return (
            np.array([x, x + self.dx, x + self.dx, x]),
            np.array([y, y, y + self.dy, y + self.dy]),
        )

    @cached_property
    def bounds(self) -> tuple[int, int, int, int]:
        return self.xmin, self.ymin, self.xmax, self.ymax

    @classmethod
    def from_centers(
        cls,
        x_centers: np.ndarray,
        y_centers: np.ndarray,
        name=None,
        crs=WGS84,
        rounding: int | None = None,
    ) -> RegularGrid:
        """Create a regular grid from the center points of the cells.

        :param x_centers: The x-coordinates of the cell centers.
        :param y_centers: The y-coordinates of the cell centers.
        :param name: The name of the grid.
        :param crs: The coordinate reference system of the grid.
        :param rounding: The number of decimal places to round the dx and dy.
            This is useful to correct floating point errors.
        """

        # Calculate the dx and dy
        dxs = np.diff(x_centers)
        dys = np.diff(y_centers)

        dx = dxs[0]
        dy = dys[0]

        if rounding is not None:
            dx = round(dx, rounding)
            dy = round(dy, rounding)

        if not np.allclose(dxs, dx) or not np.allclose(dys, dy):
            raise ValueError("The centers are not equally spaced.")

        nx = len(x_centers)
        ny = len(y_centers)

        xmin = x_centers[0] - dx / 2
        ymin = y_centers[0] - dy / 2

        if rounding is not None:
            xmin = round(xmin, rounding)
            ymin = round(ymin, rounding)

        try:
            grid = cls(
                xmin=x_centers[0] - dx / 2,
                ymin=y_centers[0] - dy / 2,
                nx=nx,
                ny=ny,
                dx=dx,
                dy=dy,
                name=name,
                crs=crs,
            )
        except Exception as e:
            raise RuntimeError(
                f"Could not create grid from centers. "
                "If there is a mismatch in coordinates, "
                "try to set `rounding` to some value."
            ) from e

        return grid


class TNOGrid(RegularGrid):
    """Contains the grid from the TNO emission inventory
    This grid is defined as a standard lat/lon coordinate system.
    The gridpoints are at the center of the cell.
    """

    def __init__(self, dataset_path, name="TNO"):
        """Open the netcdf-dataset and read the relevant grid information.

        Parameters
        ----------
        dataset_path : str
        name : str, optional
        """
        self.dataset_path = dataset_path

        with Dataset(dataset_path) as dataset:
            self.lon_var = np.array(dataset["longitude"][:])
            self.lat_var = np.array(dataset["latitude"][:])

        self.nx = len(self.lon_var)
        self.ny = len(self.lat_var)

        # The lat/lon values are the cell-centers
        self.dx = (self.lon_var[-1] - self.lon_var[0]) / (self.nx - 1)
        self.dy = (self.lat_var[-1] - self.lat_var[0]) / (self.ny - 1)

        # Compute the cell corners
        x = self.lon_var
        y = self.lat_var
        dx2 = self.dx / 2
        dy2 = self.dy / 2

        self.cell_x = np.array([x + dx2, x + dx2, x - dx2, x - dx2])
        self.cell_y = np.array([y + dy2, y - dy2, y - dy2, y + dy2])

        # by pass the regular grid __inti__ method, as variable  have been
        # initialized here
        Grid.__init__(self, name=name, crs=WGS84)

    def cell_corners(self, i, j):
        """Return the corners of the cell with indices (i,j).

        See also the docstring of Grid.cell_corners.

        Parameters
        ----------
        i : int
        j : int

        Returns
        -------
        tuple(np.array(shape=(4,), dtype=float),
              np.array(shape=(4,), dtype=float))
            Arrays containing the x and y coordinates of the corners
        """
        return self.cell_x[:, i], self.cell_y[:, j]

    @property
    def lon_range(self):
        """Return an array containing all the longitudinal points on the grid.

        Returns
        -------
        np.array(shape=(nx,), dtype=float)
        """
        return self.lon_var

    @property
    def lat_range(self):
        """Return an array containing all the latitudinal points on the grid.

        Returns
        -------
        np.array(shape=(ny,), dtype=float)
        """
        return self.lat_var


class EDGARGrid(Grid):
    """Contains the grid from the EDGAR emission inventory

    The grid is similar to the TNO grid.
    """

    def __init__(self, dataset_path, name="EDGAR"):
        """Open the netcdf-dataset and read the relevant grid information.

        Parameters
        ----------
        dataset_path : str
        name : str, optional
        """
        self.dataset_path = dataset_path

        with Dataset(dataset_path) as dataset:
            self.lon_var = np.array(dataset["lon"][:])
            self.lat_var = np.array(dataset["lat"][:])

        self.nx = len(self.lon_var)
        self.ny = len(self.lat_var)

        # The lat/lon values are the cell-centers
        self.dx = (self.lon_var[-1] - self.lon_var[0]) / (self.nx - 1)
        self.dy = (self.lat_var[-1] - self.lat_var[0]) / (self.ny - 1)

        # Compute the cell corners
        x = self.lon_var
        y = self.lat_var
        dx2 = self.dx / 2
        dy2 = self.dy / 2

        self.cell_x = np.array([x + dx2, x + dx2, x - dx2, x - dx2])
        self.cell_y = np.array([y + dy2, y - dy2, y - dy2, y + dy2])

        super().__init__(name, crs=WGS84)

    def cell_corners(self, i, j):
        """Return the corners of the cell with indices (i,j).

        See also the docstring of Grid.cell_corners.

        Parameters
        ----------
        i : int
        j : int

        Returns
        -------
        tuple(np.array(shape=(4,), dtype=float),
              np.array(shape=(4,), dtype=float))
            Arrays containing the x and y coordinates of the corners
        """
        return self.cell_x[:, i], self.cell_y[:, j]

    @cached_property
    def lon_range(self):
        """Return an array containing all the longitudinal points on the grid.

        Returns
        -------
        np.array(shape=(nx,), dtype=float)
        """
        return self.lon_var

    @cached_property
    def lat_range(self):
        """Return an array containing all the latitudinal points on the grid.

        Returns
        -------
        np.array(shape=(ny,), dtype=float)
        """
        return self.lat_var

    @cached_property
    def cell_areas(self):
        """Return an array containing the grid cell areas.

        Returns
        -------
        np.array(shape=(nx,ny), dtype=float)
        """
        lats_c = np.append(self.cell_y[1], self.cell_y[0, -1])
        lats_c = np.deg2rad(lats_c)

        dlon = 2 * np.pi / self.nx
        areas = (
            R_EARTH * R_EARTH * dlon * np.abs(np.sin(lats_c[:-1]) - np.sin(lats_c[1:]))
        )
        areas = np.broadcast_to(areas[np.newaxis, :], (self.nx, self.ny))

        return areas.flatten()


class GeoPandasGrid(Grid):
    """A grid that can be easily constructed on a geopandas dataframe."""

    def __init__(
        self,
        gdf: gpd.GeoDataFrame,
        name: str = "gpd_grid",
        shape: tuple[int, int] | None = None,
    ):
        super().__init__(name, gdf.crs)

        self._gdf = gdf

        if shape is not None:
            self.nx, self.ny = shape
        else:
            self.nx = len(gdf)
            self.ny = 1

    @property
    def cells_as_polylist(self) -> list[Polygon]:
        """Return all the cells as a list of polygons."""
        return self.gdf.geometry.tolist()


class SwissGrid(RegularGrid):
    """Represent a grid used by swiss inventories, such as meteotest, maiolica
    or carbocount."""

    dx: float
    dy: float
    xmin: float
    ymin: float

    def __init__(self, name, nx, ny, dx, dy, xmin, ymin, crs: int = LV95):
        """Store the grid information.

        Swiss grids use LV95 coordinates, which switch the axes:

        * x <-> Northing
        * y <-> Easting

        For consistency with the other Grids, we use:

        * x <-> Longitude ~ "swiss y"
        * y <-> Latitude  ~ "swiss x"

        Thus, a header of a .asc file translates as follows:

        * ncols     -> nx
        * nrows     -> ny
        * xllcorner -> ymin
        * yllcorner -> xmin
        * cellsize  -> dx, dy

        Parameters
        ----------
        dx : float
            EASTERLY size of a gridcell in meters
        dy : float
            NORTHLY size of a gridcell in meters
        nx : int
            Number of cells in EASTERLY direction
        ny : int
            Number of cells in NORTHLY direction
        xmin : float
            EASTERLY distance of bottom left gridpoint in meters
        ymin : float
            NORTHLY distance of bottom left gridpoint in meters
        crs:
            The projection to use.
            only
            work with WGS84 or  SWISS .
        """

        # The swiss grid is not technically using a PlateCarree projection
        # (in fact it is not using any projection implemented by cartopy),
        # however the points returned by the cell_corners() method are in
        # WGS84, which PlateCarree defaults to.
        super().__init__(
            name=name, nx=nx, ny=ny, dx=dx, dy=dy, xmin=xmin, ymin=ymin, crs=crs
        )

    @cached_property
    def lon_range(self):
        """Return an array containing all the longitudinal points on the grid.

        Returns
        -------
        np.array(shape=(nx,), dtype=float)
        """
        return np.array([self.xmin + i * self.dx for i in range(self.nx + 1)])

    @cached_property
    def lat_range(self):
        """Return an array containing all the latitudinal points on the grid.

        Returns
        -------
        np.array(shape=(ny,), dtype=float)
        """
        return np.array([self.ymin + j * self.dy for j in range(self.ny + 1)])


class ICONGrid(Grid):
    """Class to manage an ICON-domain

    This grid is defined as an unstuctured triangular grid (1D).
    The cells are ordered in a deliberate way and indexed with ascending integer numbers.
    The grid file contains variables like midpoint coordinates etc as a fct of the index.
    """

    def __init__(self, dataset_path, name="ICON"):
        """Open the netcdf-dataset and read the relevant grid information.

        Parameters
        ----------
        dataset_path : str
        name : str, optional
        """
        self.dataset_path = dataset_path

        with Dataset(dataset_path) as dataset:
            self.clon_var = np.rad2deg(dataset["clon"][:])
            self.clat_var = np.rad2deg(dataset["clat"][:])
            self.cell_areas = np.array(dataset["cell_area"][:])
            self.vlat = np.rad2deg(dataset["vlat"][:])
            self.vlon = np.rad2deg(dataset["vlon"][:])
            self.vertex_of_cell = np.array(dataset["vertex_of_cell"][:])
            self.cell_of_vertex = np.array(dataset["cells_of_vertex"][:])

        self.ncell = len(self.clat_var)

        # Initiate a list of polygons, which is updated whenever the polygon of a cell
        # is called for the first time
        self.polygons = [
            Polygon(zip(*self._cell_corners(i))) for i in range(self.ncell)
        ]

        # Create a geopandas df
        # ICON_FILE_CRS = 6422
        # Apparently the crs of icon is not what is written in the nc file.
        ICON_FILE_CRS = WGS84

        self._gdf = gpd.GeoDataFrame(geometry=self.polygons, crs=ICON_FILE_CRS)
        self.process_overlap_antimeridian()

        # Consider the ICON-grid as a 1-dimensional grid where ny=1
        self.nx = self.ncell
        self.ny = 1

        super().__init__(name, crs=ICON_FILE_CRS)

    def _cell_corners(self, n):
        """Internal cell corners"""

        return (
            self.vlon[self.vertex_of_cell[:, n] - 1],
            self.vlat[self.vertex_of_cell[:, n] - 1],
        )

    def cell_corners(self, n, j):
        """Return the corners of the cell with index n.

        Parameters
        ----------
        n : int
        j : int

        Returns
        -------
        tuple(np.array(shape=(3,), dtype=float),
              np.array(shape=(3,), dtype=float))
            Arrays containing the lon and lat coordinates of the corners
        """

        return self.gdf.geometry.iloc[n].exterior.coords.xy

    def gridcell_areas(self):
        """Calculate 2D array of the areas (m^2) of a regular rectangular grid
        on earth.

        Returns
        -------
        np.array
            2D array containing the areas of the gridcells in m^2
            shape: (ncell)
        """

        return self.cell_areas

    def process_overlap_antimeridian(self):
        """Find polygons intersecting the antimeridian line
        and split them into two polygons represented by a
        MultiPolygon.
        """

        def shift_lon_poly(poly):
            coords = poly.exterior.coords
            lons = np.array([coord[0] for coord in coords])
            lats = [coord[1] for coord in coords]
            if np.any(lons > 180):
                lons -= 360
            elif np.any(lons < -180):
                lons += 360
            return Polygon([*zip(lons, lats)])

        def detect_antimeridian_poly(poly):
            coords = poly.exterior.coords
            lon1, lon2, lon3 = coords[0][0], coords[1][0], coords[2][0]
            coords_cond1 = [list(c) for c in coords[:-1]]

            cond1 = np.count_nonzero(np.array([lon1, lon2, lon3]) > 180.0 - 1e-5) == 2
            if cond1:
                if lon1 < 0:
                    coords_cond1[1][0] = lon2 - 360
                    coords_cond1[2][0] = lon3 - 360
                elif lon2 < 0:
                    coords_cond1[0][0] = lon1 - 360
                    coords_cond1[2][0] = lon3 - 360
                elif lon3 < 0:
                    coords_cond1[0][0] = lon1 - 360
                    coords_cond1[1][0] = lon2 - 360

            vmin = -140
            vmax = 140
            lon1, lon2, lon3 = (
                coords_cond1[0][0],
                coords_cond1[1][0],
                coords_cond1[2][0],
            )
            cond2 = (
                (lon1 > vmax or lon1 < vmin)
                or (lon2 > vmax or lon2 < vmin)
                or (lon3 > vmax or lon3 < vmin)
            )
            coords_cond2 = [list(c) for c in coords_cond1]

            if cond2:
                if lon1 * lon2 < 0 and lon1 * lon3 < 0:
                    coords_cond2[0][0] = lon1 - math.copysign(1, lon1) * 360

                elif lon2 * lon1 < 0 and lon2 * lon3 < 0:
                    coords_cond2[1][0] = lon2 - math.copysign(1, lon2) * 360

                elif lon3 * lon1 < 0 and lon3 * lon2 < 0:
                    coords_cond2[2][0] = lon3 - math.copysign(1, lon3) * 360

            return Polygon(coords_cond2)

        crs = pyproj.CRS.from_epsg(WGS84)
        bounds = crs.area_of_use.bounds

        xx_bounds, yy_bounds = box(*bounds).exterior.coords.xy
        coords_bounds = [(x, y) for x, y in zip(xx_bounds, yy_bounds)]
        bounds_line = LineString(coords_bounds)

        self._gdf = self.gdf.set_geometry(
            self.gdf.geometry.apply(lambda poly: detect_antimeridian_poly(poly))
        )
        gdf_inter = self._gdf.loc[self.gdf.intersects(bounds_line)]
        gdf_inter = gdf_inter.set_geometry(
            gdf_inter.geometry.apply(
                lambda poly: MultiPolygon(split(poly, bounds_line))
            )
        )
        gdf_inter = gdf_inter.set_geometry(
            gdf_inter.geometry.apply(
                lambda mpoly: MultiPolygon(
                    [shift_lon_poly(poly) for poly in mpoly.geoms]
                )
            )
        )
        self._gdf.loc[gdf_inter.index, "geometry"] = gdf_inter.geometry



"""Read an the input for a gramgral simulation as an inventory."""
from __future__ import annotations
from enum import Enum, IntEnum
import json
import logging
from os import PathLike
from pathlib import Path
from typing import Any
from emiproc.grids import LV95, WGS84
from emiproc.inventories import Category, Inventory, Substance
from emiproc.exports.gral import EmissionWriter
from emiproc.utilities import HOUR_PER_YR
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point, LineString, Polygon


class PointsCols(IntEnum):
    """Columns of the point sources file."""

    X = 0
    Y = 1
    Z = 2
    EMISSION = 3  # kg/h
    EXIT_VELOCITY = 7
    STACK_DIAMETER = 8
    EXIT_TEMPERATURE = 9
    SOURCE_GROUP = 10


class LinesCols(IntEnum):
    """Columns of the line sources file."""

    NAME = 0
    SECTION = 1
    SOURCE_GROUP = 2
    X_START = 3
    Y_START = 4
    Z_START = 5
    X_END = 6
    Y_END = 7
    Z_END = 8
    WIDTH = 9
    EXTENSION = 10
    EMISSION = 13  # kg/h/km

class CadastreCols(IntEnum):
    """Columns of the cadastre file."""

    X = 0
    Y = 1
    Z = 2
    X_EXTENSION = 3
    Y_EXTENSION = 4
    Z_EXTENSION = 5
    EMISSION = 6  # kg/h
    SOURCE_GROUP = 10

class GralInventory(Inventory):
    """Gral inventory.

    Load the inputs of a gral simulation.
    
    .. warning:: There are some things not implemented yet:
        - The portals file
        - Emission Infos (exit velocity, heights, ...)
        - Shared utilites between thefiles (many code is duplicated accros funcitons)

    """

    def __init__(
        self,
        emissions_dir: PathLike,
        source_group_mapping: dict[int, tuple[Substance, Category]] | None = None,
        crs: str | int = WGS84,
    ) -> None:

        emissions_dir = Path(emissions_dir)
        if not emissions_dir.is_dir():
            raise ValueError(f"{emissions_dir=} is not a directory")
        self.path = emissions_dir

        self.file_points = self.path / "point.dat"
        self.file_lines = self.path / "line.dat"
        self.file_cadastre = self.path / "cadastre.dat"
        self.file_portals = self.path / "portals.dat"

        group_mapping_file = self.path / "source_groups.json"

        super().__init__()

        # Read the source group mapping
        if source_group_mapping is None:
            if not group_mapping_file.is_file():
                raise FileNotFoundError(
                    f"{group_mapping_file=} not found. Please provide"
                    " 'source_group_mapping' of source groups to substances and"
                    " categories. Or generate the file."
                )
            with open(group_mapping_file) as f:
                source_group_mapping = json.load(f)
                self.logger.debug(f"From {group_mapping_file}: {source_group_mapping=}")
                # Convert the keys to int
                source_group_mapping = {
                    int(k): v for k, v in source_group_mapping.items()
                }
        else:
            if group_mapping_file.is_file():
                self.logger.warning(
                    f"{group_mapping_file=} exist, but 'source_group_mapping' was"
                    " provided, so the file will be ignored."
                )

        self.source_group_mapping = source_group_mapping
        self.gdf = None
        self.gdfs = {}

        # assign the crs
        self._requested_crs = crs

        # Now we can read the files
        self._read_points()
        self._read_lines()
        self._read_cadastre()
        self._read_portals()

    def _get_sub_cat(self, source_group: int) -> tuple[Substance, Category]:
        """Get the substance and category for a source group."""
        if source_group not in self.source_group_mapping:
            raise ValueError(
                f"{source_group=} not found in {self.source_group_mapping=}"
            )
        return self.source_group_mapping[source_group]

    def _add_gdf(
        self, gdf: gpd.GeoDataFrame, substance: Substance, category: Category
    ) -> None:
        """Add a GeoDataFrame to the inventory."""
        if category not in self.gdfs:
            self.gdfs[category] = gdf
        else:
            # Add to the existing GeoDataFrame
            # Set the values on missing columns to 0
            missing_columns = set(self.gdfs[category].columns) - set(gdf.columns)
            for col in missing_columns:
                gdf[col] = 0
            # Also set the substance to 0 in the gdfs if not there
            if substance not in self.gdfs[category].columns:
                self.gdfs[category][substance] = 0
            # Now we can append
            self.gdfs[category] = pd.concat([self.gdfs[category], gdf], ignore_index=True)

    def _read_points(self) -> None:
        """Read the point sources."""
        if not self.file_points.is_file():
            self.logger.warning(f"{self.file_points=} not found.")
            return

        df = pd.read_csv(self.file_points, sep=",", header=1)
        # find all the source categories
        self.logger.debug(f"{df.columns=}")
        source_groups = df[df.columns[PointsCols.SOURCE_GROUP]].unique()

        for source_group in source_groups:
            substance, category = self._get_sub_cat(source_group)
            mask_source_group = df[df.columns[PointsCols.SOURCE_GROUP]] == source_group
            # Create the GeoDataFrame
            emission_col = df.columns[PointsCols.EMISSION]
            # Select only the emission column renamed to the substance
            emissions_values = df[mask_source_group].rename(
                columns={emission_col: substance}
            )[substance].to_numpy().astype(float)
            # Convert the units from kg/h to kg/y
            emissions_values *= HOUR_PER_YR
            gdf = gpd.GeoDataFrame(
                {substance: emissions_values},
                geometry=gpd.points_from_xy(
                    df[mask_source_group][df.columns[PointsCols.X]],
                    df[mask_source_group][df.columns[PointsCols.Y]],
                ),
                crs=self._requested_crs,
            )
            self._add_gdf(gdf, substance, category)
            # TODO: add the source information as well

    def _read_lines(self) -> None:
        """Read the line sources."""
        if not self.file_lines.is_file():
            self.logger.warning(f"{self.file_lines=} not found.")
            return

        df = pd.read_csv(self.file_lines, sep=",", header=4)
        # find all the source categories
        self.logger.debug(f"{df.columns=}")
        source_groups = df[df.columns[LinesCols.SOURCE_GROUP]].unique()

        for source_group in source_groups:
            substance, category = self._get_sub_cat(source_group)
            self.logger.debug(f"{source_group=}, {substance=}, {category=}")
            mask_source_group = df[df.columns[LinesCols.SOURCE_GROUP]] == source_group
            # Create the GeoDataFrame
            emission_col = df.columns[LinesCols.EMISSION]
            # Create geseries with the start points and end points
            gs_start = gpd.points_from_xy(
                df[mask_source_group][df.columns[LinesCols.X_START]],
                df[mask_source_group][df.columns[LinesCols.Y_START]],
            )
            gs_end = gpd.points_from_xy(
                df[mask_source_group][df.columns[LinesCols.X_END]],
                df[mask_source_group][df.columns[LinesCols.Y_END]],
            )
            # Create the linestrings
            gs_lines = gpd.GeoSeries(
                [LineString([start, end]) for start, end in zip(gs_start, gs_end)],
                crs=self._requested_crs,
            )
            self.logger.debug(f"{gs_lines=}")

            # Select only the emission column renamed to the substance
            emission_values = df[mask_source_group].rename(
                columns={emission_col: substance}
            )[substance].to_numpy()
            # Convert the units from kg/h/km to kg/y(/shape)
            line_lenghts = gs_lines.length * 1e-3
            emission_values *= HOUR_PER_YR * line_lenghts

            self.logger.debug(f"{emission_values=}")
            # Create the GeoDataFrame
            gdf = gpd.GeoDataFrame(
                {substance: emission_values},
                geometry=gs_lines,
                crs=self._requested_crs,
            )
            self.logger.debug(f"{gdf=}")
            self._add_gdf(gdf, substance, category)

    def _read_cadastre(self) -> None:
        """Read the cadastre sources."""
        if not self.file_cadastre.is_file():
            self.logger.debug(f"{self.file_cadastre=} not found.")
            return

        df = pd.read_csv(self.file_cadastre, sep=",", header=0, index_col=False)
        self.logger.debug(f"{df}")
        # find all the source categories
        self.logger.debug(f"{df[df.columns[CadastreCols.SOURCE_GROUP]]=}")
        source_groups = df[df.columns[CadastreCols.SOURCE_GROUP]].unique()
        self.logger.debug(f"{source_groups=}")
        for source_group in source_groups:
            substance, category = self._get_sub_cat(source_group)
            self.logger.debug(f"{source_group=}, {substance=}, {category=}")
            mask_source_group = df[df.columns[CadastreCols.SOURCE_GROUP]] == source_group
            # Create the GeoDataFrame
            emission_col = df.columns[CadastreCols.EMISSION]
            # Create geseries with the start points and end points
            df_group = df[mask_source_group]

            gs_sqaures = gpd.GeoSeries(
                [Polygon(((x, y), (x + x_ext, y), (x + x_ext, y + y_ext), (x, y + y_ext), (x, y))) 
                for x, y, x_ext, y_ext in zip(
                    df_group[df.columns[CadastreCols.X]],
                    df_group[df.columns[CadastreCols.Y]],
                    df_group[df.columns[CadastreCols.X_EXTENSION]],
                    df_group[df.columns[CadastreCols.Y_EXTENSION]],
                )],
                crs=self._requested_crs,
            )
            self.logger.debug(f"{gs_sqaures=}")

            # Select only the emission column renamed to the substance
            emission_values = df[mask_source_group].rename(
                columns={emission_col: substance}
            )[substance].to_numpy()
            # Convert the units from kg/h(/shape?) to kg/y(/shape)
            emission_values *= HOUR_PER_YR 

            self.logger.debug(f"{emission_values=}")
            # Create the GeoDataFrame
            gdf = gpd.GeoDataFrame(
                {substance: emission_values},
                geometry=gs_sqaures,
                crs=self._requested_crs,
            )
            self.logger.debug(f"{gdf=}")
            self._add_gdf(gdf, substance, category)



    def _read_portals(self) -> None:
        """Read the portals sources."""
        if not self.file_portals.is_file():
            self.logger.debug(f"{self.file_portals=} not found.")
            return
        else:
            self.logger.error(
                f"{self.file_portals=} found. But reading portals is not implemented"
                " yet."
            )







"""Plot the tno inventory."""

# %%
from pathlib import Path
import matplotlib.pyplot as plt
from emiproc.plots import plot_inventory
from emiproc.inventories.tno import TNO_Inventory
from emiproc.inventories.utils import group_categories
from emiproc.regrid import remap_inventory

plt.style.use("default")
# %%
# path to input inventory
input_path = Path(r"C:\Users\coli\Documents\Data\TNO\TNO_GHGco_v1_1_year2015.nc")
# output path and filename
output_dir = input_path.parent
output_path = output_dir / f"{input_path.stem}_with_emissions.nc"

# %%
inv = TNO_Inventory(
    input_path,
    # Take only CH4
    substances_mapping={"co2_ff": "CO2", "co2_bf": "CO2"},
)

# %% Rempa the point sources
remapped = remap_inventory(inv, inv.grid)

# %%
grouped = group_categories(remapped, {"total": remapped.categories})


# %%
plot_inventory(
    grouped,
    cmap="magma_r",
    total_only=True,
    figsize=(16, 7),
    vmin=1e-8,
    vmax=1e2,
)

# %%



import xarray as xr
import numpy as np


da_profiles_indexes_catsub = xr.DataArray(
    data=np.array([[0, 1, 1], [2, 0, 3]]),
    coords=[["a", "b"], ["CO2", "CH4", "NOx"]],
    dims=["category", "substance"],
)
da_profiles_indexes_sub = xr.DataArray(
    data=np.array([0, 1, 1]),
    coords=[["CO2", "CH4", "NOx"]],
    dims=["substance"],
)

if __name__ == "__main__":
    from emiproc.profiles.utils import get_desired_profile_index

    print(da_profiles_indexes_catsub)

    print(get_desired_profile_index(da_profiles_indexes_catsub, cat="a", sub="CH4"))
    print(
        get_desired_profile_index(
            da_profiles_indexes_sub,
            cat="a",
            # sub="CH4"
        )
    )
    print(
        get_desired_profile_index(
            da_profiles_indexes_catsub, cell=43, cat="a", sub="CH4"
        )
    )



"""Gets the largest point source in the inventory."""

# %%
from pathlib import Path

import pandas as pd

from emiproc.inventories.zurich import MapLuftZurich

# %% define some parameters for the output


YEAR = 2022

INCLUDE_SWISS_OUTSIDE = False
swiss_data_path = Path(
    r"C:\Users\coli\Documents\ZH-CH-emission\Data\CHEmissionen\CH_Emissions_2015_2020_2022_CO2_CO2biog_CH4_N2O_BC_AP.xlsx"
)
outdir = Path(r"C:\Users\coli\Documents\ZH-CH-emission\output_files\mapluft_rasters")
mapluft_dir = Path(r"C:\Users\coli\Documents\Data\mapluft_emissionnen_kanton")
mapluf_file = mapluft_dir / f"mapLuft_{YEAR}_v2024.gdb"


# %% load the zurich inventory
inv = MapLuftZurich(mapluf_file, substances=["CO2"])
# %%
gdf_biggest_emissions = []
for cat, gdf in inv.gdfs.items():
    # Check if the geometry is a point
    if gdf.geometry.type[0] != "Point":
        continue

    gdf_max = gdf.sort_values("CO2").tail(100)
    gdf_max["category"] = cat
    gdf_biggest_emissions.append(gdf_max)

gdf_biggest_emissions = pd.concat(gdf_biggest_emissions)
biggest_emitters = gdf_biggest_emissions.sort_values("CO2", ascending=False)
biggest_emitters
# %%
biggest_emitters["x"] = biggest_emitters.geometry.x
biggest_emitters["y"] = biggest_emitters.geometry.y
biggest_emitters.iloc[:100].to_csv("biggest_point_sources.csv")
# %%




import geopandas as gpd
from emiproc.regrid import geoserie_intersection
from shapely.geometry import  Polygon
import numpy as np


# prepare test data for the tests
serie = gpd.GeoSeries(
    [
        Polygon(((0, 0), (0, 1), (1, 1), (1, 0))),
        Polygon(((2, 1), (2, 2), (3, 2), (3, 1))),
        Polygon(((0, 1), (0, 2), (1, 2), (1, 1))),
        Polygon(((1, 0), (1, 1), (2, 1), (2, 0))),
        Polygon(((1, 1), (1, 2), (2, 2), (2, 1))),
    ]
)
triangle = Polygon(((0.5, 0.5), (1.5, 0.5), (1.5, 1.5)))

expected_weights_droped = np.array([0.125, 0.25 , 0.125])
expected_weights_undroped = np.array([0.125, 0.,  0.   , 0.25 , 0.125])


def test_normal_intersection():

    intersection, weights = geoserie_intersection(serie, triangle)

    # Shape number 1 and 2 shoud have disappeared
    assert len(intersection) == 3
    # We cannot test the follwoing sadly
    print(intersection.iloc[0])
    
    # Test the shapes are what we expect
    assert intersection.iloc[0].equals(Polygon(((0.5, 0.5), (1, 0.5), (1, 1))))
    assert intersection.iloc[1].equals(Polygon(((1, 0.5), (1.5, 0.5), (1.5, 1), (1, 1))))
    assert intersection.iloc[2].equals(Polygon(((1.5, 1.5), (1.5, 1), (1, 1))))

    assert np.all(expected_weights_droped == weights)

def test_intersection_drop_unused():

    intersection, weights = geoserie_intersection(serie, triangle, drop_unused=False)

    # Shape number 1 shoud have disappeared
    assert 1  in intersection.index
    assert np.all(expected_weights_undroped == weights)

def test_intersection_keep_outside():

    intersection, weights = geoserie_intersection(serie, triangle, keep_outside=True)

    assert np.all(expected_weights_undroped == (1-weights))




"""Different functions for doing the weights remapping."""

from __future__ import annotations
import logging
from pathlib import Path
from warnings import warn
import numpy as np
import geopandas as gpd
from typing import TYPE_CHECKING, Iterable
from shapely.geometry import Point, MultiPolygon, Polygon
from emiproc.utilities import ProgressIndicator
from scipy.sparse import coo_array, dok_matrix
from emiproc.grids import Grid
from emiproc.profiles.operators import get_weights_of_gdf_profiles, remap_profiles

logger = logging.getLogger("emiproc.regrid")

if TYPE_CHECKING:
    from os import PathLike
    from emiproc.inventories import Inventory


def get_weights_mapping(
    weights_filepath: Path | None,
    shapes_inv: Iterable[Polygon | Point],
    shapes_out: Iterable[Polygon],
    loop_over_inv_objects: bool = False,
    method: str = "new",
) -> dict[str, np.ndarray]:
    """Get the requested weights mapping.

    If it does not exists, calls :py:func:`calculate_weights_mapping` .
    See that function for the other arguments.
    and save the weights once computed.

    :arg weights_filepath: The name of the file in which to save the
        weights data. Emiproc will add some metadata to it.
        This file has to be a npz archive ending with suffix .npz .
        Emiproc will add the suffix if you don't.
    :arg shapes_inv: The shapes of the inventory.
        Shapes from which the remapping will be done.
    :arg shapes_out: The shapes to which the remapping will be done.
    :arg loop_over_inv_objects: Whether the loop should happend on the
        the inventory objects instead of the output shapes.
        This will be where the performance bottleneck resides.
        It is difficult to guess what is the good option but it
        can make big differences in some cases.
        If you have point sources in your shapes_inv, this MUST be set
        to True.


    """
    logger.debug(
        "get_weights_mapping("
        f"{weights_filepath=},"
        f"{shapes_inv=},"
        f"{shapes_out=},"
        f"{loop_over_inv_objects=},"
        f"{method=},"
        ")"
    )
    if weights_filepath is not None:
        weights_filepath = Path(weights_filepath).with_suffix(f".npz")
        if loop_over_inv_objects:
            # Add a small marker
            weights_filepath = weights_filepath.with_stem(
                weights_filepath.stem + "_loopinv"
            )

    if (weights_filepath is None) or (not weights_filepath.exists()):
        w_mapping = calculate_weights_mapping(
            shapes_inv, shapes_out, loop_over_inv_objects, method
        )
        if weights_filepath is not None:
            # Make sure dir is created
            weights_filepath.parent.mkdir(exist_ok=True, parents=True)
            np.savez(weights_filepath, **w_mapping)

    else:
        w_mapping = {**np.load(weights_filepath)}

    return w_mapping


def calculate_weights_mapping(
    shapes_inv: Iterable[Polygon | Point | MultiPolygon],
    shapes_out: Iterable[Polygon],
    loop_over_inv_objects: bool = False,
    method: str = "new",
) -> dict[str, np.ndarray]:
    """Return a dictionary with the mapping.

    Every weight means: From which shape in the invetory
    to which shape in the output and the weight value is the proportion
    of the inventory weight present in the output.

    The output contains the weigths mapping.

    Point source ending in more than one cell will be splitt among the cells.

    :arg inv_indexes: The indexes from which shape in the inverntory
    :arg output_indexes: The indexes from which shape in the output
    :arg weights: The weight of this connexion (between 0 and 1).
        It means the percentage of the inv shape that should go in
        the output.
    """

    # shapes_inv = inv.gdf.geometry
    # shapes_out = grid.gdf.to_crs(inv.crs)
    # loop_over_inv_objects=False
    logger.info(
        "calculating weights mapping "
        f"from {len(shapes_inv)} inventory shapes "
        f"to {len(shapes_out)} grid cells."
    )

    w_mapping = {
        "inv_indexes": [],
        "output_indexes": [],
        "weights": [],
    }

    if loop_over_inv_objects:
        shapes_vectorized = shapes_out
        shapes_looped = shapes_inv
    else:
        shapes_vectorized = shapes_inv
        shapes_looped = shapes_out

    if isinstance(shapes_vectorized, gpd.GeoDataFrame):
        shapes_vect = shapes_vectorized.geometry
    elif isinstance(shapes_vectorized, gpd.GeoSeries):
        shapes_vect = shapes_vectorized
    elif isinstance(shapes_vectorized, list):
        shapes_vect = gpd.GeoSeries(shapes_vectorized)
    else:
        raise TypeError(f"'shapes_vectorized' cannot be {type(shapes_vectorized)}")
    shapes_vect: gpd.GeoSeries

    # Check that only area sources are vectorized
    if not np.all(
        shapes_vect.map(lambda shape: isinstance(shape, (Polygon, MultiPolygon)))
    ):
        raise TypeError(
            "Non Polygon geometries were found on the grid but cannot be used for"
            " remapping. Use 'loop_over_inv_objects' if you want to remap points to a"
            " grid."
        )

    if isinstance(shapes_looped, gpd.GeoDataFrame):
        shapes_looped = shapes_looped.geometry
    elif isinstance(shapes_looped, gpd.GeoSeries):
        shapes_looped = shapes_looped
    elif isinstance(shapes_looped, list):
        shapes_looped = gpd.GeoSeries(shapes_looped)
    else:
        raise TypeError(f"'shapes_looped' cannot be {type(shapes_looped)}")
    shapes_looped: gpd.GeoSeries
    minx, miny, maxx, maxy = shapes_looped.total_bounds
    if minx != maxx and miny != maxy:
        # Seems to remove all the data if boundaries are equal
        # Mask with only what is in the bounds
        shapes_vect = shapes_vect.cx[minx:maxx, miny:maxy]

    progress = ProgressIndicator(len(shapes_looped))

    if method == "old":
        # Loop over the output shapes
        for looped_index, shape in enumerate(shapes_looped):
            progress.step()

            intersect = shapes_vect.intersects(shape)
            if np.any(intersect):
                # Find where the intesection occurs
                intersecting_serie = shapes_vect.loc[intersect]
                from_indexes = intersecting_serie.index.to_list()
                if isinstance(shape, Point):
                    # Check in which areas the point ended
                    n_areas = len(from_indexes)
                    weights = np.full(n_areas, 1.0 / n_areas)

                else:
                    # Calculate the intersection areas
                    areas = (
                        intersecting_serie.intersection(shape)
                        .area.to_numpy()
                        .reshape(-1)
                    )
                    if loop_over_inv_objects:
                        # Take the ratio of the shape from the overlap from the shape i
                        weights = areas / shape.area
                    else:
                        # Take ratio of all the inventory shapes that were crossed
                        weights = areas / intersecting_serie.area

                # Find out the mapping indexes
                looped_indexes = np.full_like(from_indexes, looped_index)
                w_mapping["output_indexes"].append(
                    from_indexes if loop_over_inv_objects else looped_indexes
                )

                w_mapping["inv_indexes"].append(
                    looped_indexes if loop_over_inv_objects else from_indexes
                )

                w_mapping["weights"].append(weights)

        for key, l in w_mapping.items():
            if l:  # If any weights were added
                w_mapping[key] = np.concatenate(l, axis=0).reshape(-1)

    elif method == "new":
        # Merge the two geometries using intersections
        gdf_in = gpd.GeoDataFrame(geometry=shapes_vect)
        gdf_out = gpd.GeoDataFrame(geometry=shapes_looped)
        gdf_weights = gdf_in.sjoin(gdf_out, rsuffix="out")
        gdf_weights = gdf_weights.merge(
            gdf_out, left_on="index_out", right_index=True, suffixes=("", "_out")
        )
        gdf_weights.index.name = "index_inv"
        gdf_weights = gdf_weights.assign(
            geometry_inter=lambda d: (
                d["geometry"].intersection(gpd.GeoSeries(d["geometry_out"]))
            )
        )

        if loop_over_inv_objects:
            # Calculate weights for polygons
            gdf_weights["weights"] = (
                gdf_weights.geometry_inter.area / gdf_weights.geometry_out.area
            )

            # Process the points
            gdf_points = gdf_weights.loc[gdf_weights.geometry_out.type == "Point"]
            if gdf_points.shape[0]:
                nareas_points = gdf_points.groupby("index_out").transform(
                    np.count_nonzero
                )["geometry"]
                gdf_weights.loc[gdf_weights.geometry_out.type == "Point", "weights"] = (
                    1 / nareas_points
                )

            # Extract indices
            gdf_weights = gdf_weights.sort_values(by=["index_inv", "index_out"])
            w_mapping["inv_indexes"] = gdf_weights.index_out.to_numpy()
            w_mapping["output_indexes"] = gdf_weights.index.to_numpy()

        else:
            # Calculate weights and extract indices
            gdf_weights["weights"] = (
                gdf_weights.geometry_inter.area / gdf_weights.geometry.area
            )
            gdf_weights = gdf_weights.sort_values(by=["index_out", "index_inv"])
            w_mapping["inv_indexes"] = gdf_weights.index.to_numpy()
            w_mapping["output_indexes"] = gdf_weights.index_out.to_numpy()

        w_mapping["weights"] = gdf_weights.weights.to_numpy()

    else:
        raise ValueError(f"'method' must be one of ['new', 'old'] not {method}.")
    # Ensure types
    w_mapping["output_indexes"] = np.array(w_mapping["output_indexes"], dtype=int)
    w_mapping["inv_indexes"] = np.array(w_mapping["inv_indexes"], dtype=int)
    w_mapping["weights"] = np.array(w_mapping["weights"], dtype=float)

    return w_mapping


def weights_remap(
    w_mapping: dict[str, np.ndarray],
    remapped_values: np.ndarray,
    output_size: int | tuple,
) -> np.ndarray:
    """Remap using the weights mapping and a sparse matrix.

    This allows for a very fast dot product calculation if the mapping.
    Is sparse.
    """
    if isinstance(output_size, int):
        out_len = output_size
    else:
        out_len = np.prod(output_size)

    A = coo_array(
        (w_mapping["weights"], (w_mapping["output_indexes"], w_mapping["inv_indexes"])),
        shape=(out_len, len(remapped_values)),
        dtype=float,
    )

    return A.dot(remapped_values).reshape(output_size)


def weights_remap_matrix(
    w_matrix: coo_array,
    remapped_values: np.ndarray,
) -> np.ndarray:
    """Remap using the weights mapping and a sparse matrix.

    This is the same as :py:func:`weights_remap` but uses a matrix as input.
    This allow for not having to build the matrix multiple times
    """
    return w_matrix.dot(remapped_values)


def geoserie_intersection(
    geometry: gpd.GeoSeries,
    shape: Polygon,
    keep_outside: bool = False,
    drop_unused: bool = True,
) -> tuple[gpd.GeoSeries, np.ndarray]:
    """Calculate the intersection of a geoserie and a shape.

    This can be an expensive operation you might want to cache.

    .. note:: This is used for cropping polygons only.

    :arg geometry: The serie of shapes from you inventory or grid.
    :arg shape: A polygon which will be used for cropping.
    :arg keep_outside: Whether to keep only the outer region of the geometry
        instead the of the inner.
        If this is true, the weigth
    :arg drop_unused: Whether all the shapes from the geometry serie should
        be kept. If True, the returned serie will remove these grid shapes.
        The index of the returned serie will correspond to the original shapes.
        If False, the returned geoserie will contain geometries from the
        original grid but the weights will be 0.

    :return cropped_shapes, weights: Return a tuple containing
        the shapes of the data but
        cropped with respect to the shape and
        the weights that correspond to the proportion of the original
        geometry shapes that is present in the cropping shape.
        The weights are 0 if the geometry is not included in the cropping shape
        and 1 if the geometry is fully included. This is the opposite
        when keep_outside is set to True.


    """
    # Check the geometry that intersect the shape
    mask_intersect = geometry.intersects(shape)
    mask_within = geometry.within(shape)
    mask_boundary_intersect = mask_intersect & (~mask_within)

    # Find the intersection sahpes of the boundary shapes
    shapes_boundary_intersect = geometry.loc[mask_boundary_intersect].intersection(
        shape
    )
    weigths_boundary_intersect = (
        shapes_boundary_intersect.area / geometry.loc[mask_boundary_intersect].area
    )

    weights = np.zeros(len(geometry))
    weights[mask_within] = 1.0
    weights[mask_boundary_intersect] = weigths_boundary_intersect
    intersection_shapes = geometry.copy()

    if keep_outside:
        # Outside inverses the weights
        weights = 1.0 - weights
        mask = (~mask_within) | mask_boundary_intersect
        intersection_shapes.loc[mask_boundary_intersect] = intersection_shapes.loc[
            mask_boundary_intersect
        ].difference(shape)
    else:
        mask = mask_within | mask_boundary_intersect
        intersection_shapes.loc[mask_boundary_intersect] = shapes_boundary_intersect

    # Return only what was used
    if drop_unused:
        mask = mask & (weights > 0)
        # Reset the index to make sure we just created a new grid
        return intersection_shapes.loc[mask], weights[mask]
    else:
        return intersection_shapes, weights


def remap_inventory(
    inv: Inventory,
    grid: Grid | gpd.GeoSeries,
    weigths_file: PathLike | None = None,
    method: str = "new",
    keep_gdfs: bool = False,
) -> Inventory:
    """Remap any inventory on the desired grid.

    This will also remap the additional gdfs of the inventory on that grid.


    :arg inv: The inventory from which to remap.
    :arg grid: The grid to remap to.
    :arg weigths_file: The file storing the weights.
    :arg method: The method to use for remapping. See :py:func:`calculate_weights_mapping`.
    :arg keep_gdfs: Whether to keep the additional gdfs (shapped emissions) of the inventory.

    .. warning::

        To make sure the grid is defined on the same crs as the inventory,
        this funciton will call geopandas.to_crs to the grid geometries.



    """
    if weigths_file is not None:
        weigths_file = Path(weigths_file)

    if isinstance(grid, Grid) or issubclass(type(grid), Grid):
        grid_cells = gpd.GeoSeries(grid.cells_as_polylist, crs=grid.crs)
    elif isinstance(grid, gpd.GeoSeries):
        grid_cells = grid.reset_index(drop=True)
    else:
        raise TypeError(f"grid must be of type Grid or gpd.Geoseries, not {type(grid)}")

    # Treat possible issues with crs not matching
    if inv.crs is not None:
        if grid_cells.crs != inv.crs:
            # convert the grid cells to the correct crs
            grid_cells = grid_cells.to_crs(inv.crs)
    else:
        if grid_cells.crs is not None:
            raise ValueError(
                "The inventory given has no crs, but the grid has. "
                "Assign a crs to the inventory before remapping."
            )

    if inv.gdf is not None:
        # Remap the main data
        w_mapping_grid = get_weights_mapping(
            weigths_file,
            inv.gdf.geometry,
            grid_cells,
            loop_over_inv_objects=False,
            method=method,
        )
        # Create the weights matrix
        if max(w_mapping_grid["output_indexes"]) > len(grid_cells):
            raise ValueError(
                f"Error in weights mapping: {max(w_mapping_grid['output_indexes'])=} >"
                f" {len(grid_cells)=}"
            )
        if max(w_mapping_grid["inv_indexes"]) > len(inv.gdf):
            raise ValueError(
                f"Error in weights mapping: {max(w_mapping_grid['inv_indexes'])=} >"
                f" {len(inv.gdf)=}"
            )
        w_matrix = coo_array(
            (
                w_mapping_grid["weights"],
                (w_mapping_grid["output_indexes"], w_mapping_grid["inv_indexes"]),
            ),
            shape=(len(grid_cells), len(inv.gdf)),
            dtype=float,
        )
        # Perform the remapping on each column
        mapping_dict = {
            key: weights_remap_matrix(w_matrix, inv.gdf[key])
            for key in inv._gdf_columns
        }

    else:
        mapping_dict = {}

    # Add the other mappings
    if not keep_gdfs:
        for category, gdf in inv.gdfs.items():
            # Get the weights of that gdf
            if weigths_file is None:
                w_file = None
            else:
                w_file = weigths_file.with_stem(weigths_file.stem + f"_gdfs_{category}")
            w_mapping = get_weights_mapping(
                w_file,
                gdf.geometry.reset_index(drop=True),
                grid_cells,
                loop_over_inv_objects=True,
                method=method,
            )
            # Remap each substance
            for sub in gdf.columns:
                if isinstance(gdf[sub].dtype, gpd.array.GeometryDtype):
                    continue  # Geometric column
                remapped = weights_remap(
                    w_mapping,
                    # Reset the index, same as the grid was applied in the weights mapping function
                    gdf[sub].reset_index(drop=True),
                    len(grid_cells),
                )
                if (category, sub) not in mapping_dict:
                    # Create new entry
                    mapping_dict[(category, sub)] = remapped
                else:
                    # Add it to the category
                    mapping_dict[(category, sub)] += remapped

    # Create the output inv
    out_inv = inv.copy(
        no_gdfs=True,
        # Copy the profiles to the new inventory
        profiles=True,
    )
    out_inv.grid = grid
    out_inv.gdf = gpd.GeoDataFrame(
        mapping_dict,
        geometry=grid_cells,
        crs=inv.crs,
    )
    if keep_gdfs:
        out_inv.gdfs = {key: gdf.copy(deep=True) for key, gdf in inv.gdfs.items()}
    else:
        out_inv.gdfs = {}

    # Remap the profiles as well
    for index_name, profile_name in [
        ("t_profiles_indexes", "t_profiles_groups"),
        ("v_profiles_indexes", "v_profiles"),
    ]:
        if not hasattr(inv, index_name):
            continue
        indexes = getattr(inv, index_name)
        if indexes is None or "cell" not in indexes.dims:
            continue
        profiles = getattr(inv, profile_name)

        new_profiles, new_indexes = remap_profiles(
            profiles=profiles,
            profiles_indexes=indexes,
            emissions_weights=get_weights_of_gdf_profiles(
                inv.gdf, profiles_indexes=indexes
            ),
            weights_mapping=w_mapping_grid,
        )

        out_inv.set_profiles(new_profiles, new_indexes)

    out_inv.history.append(f"Remapped to grid {grid}, {keep_gdfs=}")

    return out_inv



"""Test the weights mapping function."""
from __future__ import annotations

import pytest
import geopandas as gpd
from typing import Any, Iterable
from shapely.geometry import Point, Polygon
from emiproc.regrid import calculate_weights_mapping


# Create the geometetries of an inventory
squares = gpd.GeoSeries(
    [
        Polygon(((0, 0), (0, 1), (1, 1), (1, 0))),
        Polygon(((0, 1), (0, 2), (1, 2), (1, 1))),
        Polygon(((1, 0), (1, 1), (2, 1), (2, 0))),
        Polygon(((1, 1), (1, 2), (2, 2), (2, 1))),
        Polygon(((2, 1), (2, 2), (3, 2), (3, 1))),
    ]
)
triangles = gpd.GeoSeries(
    [
        Polygon(((0.5, 0.5), (0.5, 1.5), (1.5, 1.5))),
        Polygon(((0.5, 0.5), (1.5, 0.5), (1.5, 1.5))),
        Polygon(((2.5, 0.5), (1.5, 1.5), (1.5, 0.5))),
        Polygon(((2.5, 0.5), (2.5, 1.5), (1.5, 1.5))),
    ]
)

points = gpd.GeoSeries(
    [
        Point(0.75, 0.75),  # on the boundary of the triangles
        Point(0.25, 0.25),
        Point(1.2, 1),
        Point(1, 1),  # Between all the squares
        Point(-1, -1),  # Outside
    ]
)

expected_weights = [
    # This are the weights that should be expected
    (0, 0, 1 / 8),
    (1, 0, 1 / 4),
    (2, 0, 0),
    (3, 0, 1 / 8),
    (4, 0, 0),
    (0, 1, 1 / 8),
    (1, 1, 0),
    (2, 1, 1 / 4),
    (3, 1, 1 / 8),
    (4, 1, 0),
    (0, 2, 0),
    (1, 2, 0),
    (2, 2, 1 / 4),
    (3, 2, 1 / 8),
    (4, 2, 0),
    (0, 3, 0),
    (1, 3, 0),
    (2, 3, 0),
    (3, 3, 1 / 8),
    (4, 3, 1 / 4),
]

weights_triangle_to_square = [
    (0, 0, 0.25),
    (0, 1, 0.5),
    (0, 2, 0.0),
    (0, 3, 0.25),
    (0, 4, 0.0),
    (1, 0, 0.25),
    (1, 1, 0.0),
    (1, 2, 0.5),
    (1, 3, 0.25),
    (1, 4, 0.0),
    (2, 2, 0.5),
    (2, 3, 0.25),
    (2, 4, 0.0),
    (3, 2, 0.0),
    (3, 3, 0.25),
    (3, 4, 0.5),
]

weights_points_to_square = [
    (0, 0, 1),
    (1, 0, 1),
    (2, 2, 0.5),
    (2, 3, 0.5),
    (3, 0, 0.25),
    (3, 1, 0.25),
    (3, 2, 0.25),
    (3, 3, 0.25),
    # Point 4 is not included in the grid
]

weights_points_to_triangles = [
    (0, 0, 0.5),
    (0, 1, 0.5),
    (2, 1, 1),
    (3, 0, 0.5),
    (3, 1, 0.5),
]


def check_equal_to_weights(
    weights_tested: dict[str, Iterable[float | int]],
    weights_ref: list[tuple[int, int, float]],
):
    # We will remove the weights at each encounter
    missing_weights = weights_ref.copy()
    for f, t, w in zip(
        weights_tested["inv_indexes"],
        weights_tested["output_indexes"],
        weights_tested["weights"],
    ):
        w_tuple = (f, t, w)
        if w_tuple not in missing_weights:
            raise ValueError(f"Unexpected weight detected: {w_tuple}")
        missing_weights.remove(w_tuple)

    for missing_w_tuple in missing_weights:
        # Check weights that were added but should not
        if missing_w_tuple[2] != 0:
            raise ValueError(f"Extra weights detected: {missing_w_tuple}")


# test functions
def test_simple_case():
    check_equal_to_weights(
        calculate_weights_mapping(squares, triangles), expected_weights
    )


def test_loop_inv():
    check_equal_to_weights(
        calculate_weights_mapping(squares, triangles, loop_over_inv_objects=True),
        expected_weights,
    )


def test_simple_case_tri_to_square():
    check_equal_to_weights(
        calculate_weights_mapping(
            triangles,
            squares,
        ),
        weights_triangle_to_square,
    )


def test_loop_tri_to_square():
    check_equal_to_weights(
        calculate_weights_mapping(triangles, squares, loop_over_inv_objects=True),
        weights_triangle_to_square,
    )


def test_points():
    check_equal_to_weights(
        calculate_weights_mapping(points, squares, loop_over_inv_objects=True),
        weights_points_to_square,
    )


def test_points_on_triangles():
    check_equal_to_weights(
        calculate_weights_mapping(points, triangles, loop_over_inv_objects=True),
        weights_points_to_triangles,
    )


def test_points_not_vect_raise_error():
    with pytest.raises(TypeError):
        check_equal_to_weights(
            calculate_weights_mapping(points, squares, loop_over_inv_objects=False),
            weights_points_to_square,
        )



"""Test the calucation of total emissions."""
from emiproc.tests_utils.test_inventories import inv, inv_with_pnt_sources


from emiproc.inventories.utils import get_total_emissions


total_emissions = get_total_emissions(inv)


def test_total_emissions_values():

    assert total_emissions["NH3"]["test"] == 15
    assert total_emissions["NH3"]["__total__"] == 15


def test_categories_with_no_emissions_are_not_in_dict():
    assert "adf" not in total_emissions["NH3"]


total_emissions2 = get_total_emissions(inv_with_pnt_sources)


def test_total_emissions_values():

    assert total_emissions2["AITS"]["other"] == 3
    assert total_emissions2["NH3"]["test"] == 15
    assert total_emissions2["NH3"]["__total__"] == 15
    assert total_emissions2["CO2"]["blek"] == 6
    assert total_emissions2["CO2"]["__total__"] == 6 + 10 + 3 + 10


def test_categories_with_no_emissions_are_not_in_dict():
    assert "adf" not in total_emissions2["NH3"]



from __future__ import annotations
import geopandas as gpd
import xarray as xr

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from emiproc.inventories import Inventory


def inv_to_xarray(inv: Inventory) -> xr.DataArray:
    """Convert the total emissions of an inventory to an xarray.

    :arg inv: The inventory to convert.

    :return array: A xarray containing the emissions at each coordinates:
        (substance, category, cell).
    """
    n_cells = len(inv.gdf)
    substances = inv.substances
    categories = inv.categories
    out_array = xr.DataArray(
        data=0.0,
        coords=dict(
            substance=substances,
            category=categories,
            cell=range(n_cells),
        ),
        dims=["substance", "category", "cell"],
    )

    if inv.gdfs:
        raise ValueError("The inventory cannot contain gdfs. Please remap it first.")

    for col in inv.gdf.columns:
        if isinstance(inv.gdf[col].dtype, gpd.array.GeometryDtype):
            continue
        cat, sub = col

        # Add this column to the array
        serie = inv.gdf.loc[:, col]
        out_array.loc[dict(category=cat, substance=sub)] = serie

    out_array.attrs["units"] = "kg/year/cell"
    out_array.attrs["description"] = f"Emissions of {inv.name}"
    out_array.attrs["history"] = inv.history + ["Converted to xarray."]
    out_array.attrs["source"] = "Created by emiproc.inv_to_xarray "

    return out_array
